import functools


def start_generator(func):
    """
    A decorator to start a generator and advance it to
    the first occurrence of yield. Decorator code is
    almost for verbatim taken from David Beazley.

    Source: http://dabeaz.com/coroutines/coroutine.py
    """

    @functools.wraps(func)  # To preserve the docstring of the wrapped generator.
    def start(*args, **kwargs):
        cr = func(*args, **kwargs)
        next(cr)
        return cr

    return start



"""
The Imagers class handles multiple Imager objects and coordinates plotting and animating multiple
fisheye lens and mapped images. The mapped images are also called mosaics.
"""

from typing import Tuple, List, Union, Generator
from collections import namedtuple
import pathlib
from datetime import datetime, timedelta
import shutil

import numpy as np
import matplotlib.pyplot as plt

from asilib.imager import Imager, _haversine
import asilib.map
import asilib.utils


class Imagers:
    """
    Plot and animate multiple :py:meth:`~asilib.imager.Imager` objects.

    .. warning::

        This class is in development and not all methods are implemented/matured.

    Parameters
    ----------
    imagers: Tuple
        The Imager objects to plot and animate. 
    iter_tol: float
        The allowable time tolerance, in units of time_tol*imager_cadence, for imagers to be 
        considered synchronized. Adjusting this kwarg is useful if the imager has missing 
        data and you need to animate a mosaic.
    """
    def __init__(self, imagers:Tuple[Imager], iter_tol:float=2) -> None:
        self.imagers = imagers
        # Wrap self.imagers in a tuple if the user passes in a single Imager object.
        if isinstance(self.imagers, Imager):
            self.imagers = (self.imagers, )
        self.iter_tol = iter_tol
        return
    
    def plot_fisheye(self, ax:Tuple[plt.Axes], **kwargs):
        """
        Plots one fisheye image in each subplot, oriented with North on the top, and East on the 
        left of the image.

        Parameters
        ----------
        ax: Tuple[plt.Axes]
            Subplots corresponding to each fisheye lens image.
        kwargs: dict
            Keyword arguments directly passed into each :py:meth:`~asilib.Imager.plot_fisheye`
            method.

        Example
        -------
        >>> from datetime import datetime
        >>> 
        >>> import matplotlib.pyplot as plt
        >>> import asilib
        >>> import asilib.asi
        >>> 
        >>> time = datetime(2007, 3, 13, 5, 8, 45)
        >>> location_codes = ['FSIM', 'ATHA', 'TPAS', 'SNKQ']
        >>> 
        >>> fig, ax = plt.subplots(1, len(location_codes), figsize=(12, 3.5))
        >>> 
        >>> _imagers = []
        >>> 
        >>> for location_code in location_codes:
        >>>     _imagers.append(asilib.asi.themis(location_code, time=time))
        >>> 
        >>> for ax_i in ax:
        >>>     ax_i.axis('off')
        >>> 
        >>> asis = asilib.Imagers(_imagers)
        >>> asis.plot_fisheye(ax=ax)
        >>> 
        >>> plt.suptitle('Donovan et al. 2008 | First breakup of an auroral arc')
        >>> plt.tight_layout()
        >>> plt.show()
        """
        if not isinstance(ax, (list, tuple, np.ndarray)):
            ax = (ax,)
        assert len(ax) == len(self.imagers), 'Number of subplots must equal the number of imagers.'

        for ax_i, imager_i in zip(ax, self.imagers):
            imager_i.plot_fisheye(ax=ax_i, **kwargs)
        return
    
    def plot_map(self, overlap=False, **kwargs):
        """
        Projects multiple ASI images onto a map at an altitude that is defined in the skymap 
        calibration file.

        Parameters
        ----------
        overlap: bool
            If True, pixels that overlap between imager FOV's are overplotted such that only the 
            final imager's pixels are shown.
        kwargs: dict
            Keyword arguments directly passed into each :py:meth:`~asilib.Imager.plot_map`
            method.

        Example
        -------
        >>> from datetime import datetime
        >>> 
        >>> import matplotlib.pyplot as plt
        >>> 
        >>> import asilib
        >>> import asilib.map
        >>> import asilib.asi
        >>> 
        >>> time = datetime(2007, 3, 13, 5, 8, 45)
        >>> location_codes = ['FSIM', 'ATHA', 'TPAS', 'SNKQ']
        >>> map_alt = 110
        >>> min_elevation = 2
        >>> 
        >>> ax = asilib.map.create_map(lon_bounds=(-140, -60), lat_bounds=(40, 82))
        >>> 
        >>> _imagers = []
        >>> 
        >>> for location_code in location_codes:
        >>>     _imagers.append(asilib.asi.themis(location_code, time=time, alt=map_alt))
        >>> 
        >>> asis = asilib.Imagers(_imagers)
        >>> asis.plot_map(ax=ax, overlap=False, min_elevation=min_elevation)
        >>> 
        >>> ax.set_title('Donovan et al. 2008 | First breakup of an auroral arc')
        >>> plt.show()
        """
        if not overlap:
            self._calc_overlap_mask()  # TODO: Put into a context manager.

        for imager in self.imagers:
            imager.plot_map(**kwargs)
        return
    
    # def animate_fisheye(self):
    #     raise NotImplementedError
    
    # def animate_fisheye_gen(self):
    #     raise NotImplementedError
    
    def animate_map(self, **kwargs):
        """
        Animate an ASI mosaic. It is a wrapper for the 
        :py:meth:`~asilib.Imagers.animate_map_gen` method.

        See :py:meth:`~asilib.Imagers.animate_map_gen` documentation for the complete 
        list of kwargs.

        Example
        -------
        .. code-block:: python

            >>> import asilib
            >>> import asilib.asi
            >>> 
            >>> time_range = ('2021-11-04T06:55', '2021-11-04T07:05')
            >>> asis = asilib.Imagers(
            >>>     [asilib.asi.trex_rgb(location_code, time_range=time_range) 
            >>>     for location_code in ['LUCK', 'PINA', 'GILL', 'RABB']]
            >>>     )
            >>> asis.animate_map(lon_bounds=(-115, -85), lat_bounds=(43, 63), overwrite=True)
        """
        for _ in self.animate_map_gen(**kwargs):
            pass
        return
    
    def animate_map_gen(
        self,
        overlap=False,
        lon_bounds: tuple = (-160, -50),
        lat_bounds: tuple = (40, 82),
        ax: Union[plt.Axes, tuple] = None,
        coast_color: str = 'k',
        land_color: str = 'g',
        ocean_color: str = 'w',
        color_map: str = None,
        color_bounds: List[float] = None,
        color_norm: str = None,
        color_brighten: bool = True,
        min_elevation: float = 10,
        pcolormesh_kwargs: dict = {},
        asi_label: bool = True,
        movie_container: str = 'mp4',
        animation_save_dir: Union[pathlib.Path, str]=None,
        ffmpeg_params={},
        overwrite: bool = False,
        ) -> Generator[
            Tuple[datetime, list[datetime], list[np.ndarray], plt.Axes], None, None
        ]:
        """
        Animate an ASI mosaic.

        Parameters
        ----------
        overlap: bool
            If True, pixels that overlap between imager FOV's are overplotted such that only the 
            final imager's pixels are shown.
        lon_bounds: tuple
            The map's longitude bounds.
        lat_bounds: tuple
            The map's latitude bounds.
        ax: plt.Axes, tuple
            The subplot to put the map on. If cartopy is installed, ```ax``` must be
            a two element tuple specifying the ``plt.Figure`` object and subplot position
            passed directly as ``args`` into
            `fig.add_subplot() <https://matplotlib.org/stable/api/figure_api.html#matplotlib.figure.Figure.add_subplot>`_.
        coast_color: str
            The coast color. If None will not draw it.
        land_color: str
            The land color. If None will not draw it.
        ocean_color: str
            The ocean color. If None will not draw it.
        color_map: str
            The matplotlib colormap to use. See `matplotlib colormaps <https://matplotlib.org/stable/tutorials/colors/colormaps.html>`_
            for supported colormaps.
        color_bounds: List[float]
            The lower and upper values of the color scale. The default is: low=1st_quartile and
            high=min(3rd_quartile, 10*1st_quartile). This range works well for most cases.
        color_norm: str
            Set the 'lin' (linear) or 'log' (logarithmic) color normalization. If color_norm=None,
            the color normalization will be taken from the ASI array (if specified), and if not
            specified it will default to logarithmic.
        min_elevation: float
            Masks the pixels below min_elevation degrees.
        pcolormesh_kwargs: dict
            A dictionary of keyword arguments (kwargs) to pass directly into
            plt.pcolormesh.
        asi_label: bool
            Annotates the map with the ASI code in the center of the mapped image.
        movie_container: str
            The movie container: mp4 has better compression but avi was determined
            to be the official container for preserving digital video by the
            National Archives and Records Administration.
        ffmpeg_params: dict
            The additional/overwitten ffmpeg output parameters. The default parameters are:
            framerate=10, crf=25, vcodec=libx264, pix_fmt=yuv420p, preset=slower.
        overwrite: bool
            Overwrite the animation. If False, ffmpeg will prompt you to answer y/n if the
            animation already exists.

        Yields
        ------
        datetime.datetime
            The guide time used to keep the images synchronized.
        List[datetime.datetime]
            Nearest imager time stamps to the guide time. If the difference between
            the imager time and the guide time is greater than time_tol*imager_cadence, 
            or the imager is off, the imager is considered unsynchronized and the 
            returned time is datetime.min.
        List[np.ndarray]
            The images corresponding to the times returned above. If the that imager is
            unsynchronized, the corresponding image value is None.
        plt.Axes
            The subplot object.

        Example
        -------
        .. code-block:: python

            >>> # Animate a TREX-RGB mosaic and print the individual time stamps
            >>> # to confirm that the imagers are synchronized.
            >>> import asilib
            >>> import asilib.asi
            >>> 
            >>> time_range = ('2021-11-04T06:55', '2021-11-04T07:05')
            >>> asis = asilib.Imagers(
            >>>     [asilib.asi.trex_rgb(location_code, time_range=time_range) 
            >>>     for location_code in ['LUCK', 'PINA', 'GILL', 'RABB']]
            >>>     )
            >>> gen = asis.animate_map_gen(
            >>>     lon_bounds=(-115, -85), lat_bounds=(43, 63), overwrite=True
            >>>     )
            >>> for guide_time, asi_times, asi_images, ax in gen:
            >>>     if '_text_obj' in locals():
            >>>         _text_obj.remove()
            >>>     info_str = f'Guide: {guide_time: %Y:%m:%d %H:%M:%S}\\n'
            >>>     # The below for loop is possible because the imagers and 
            >>>     # asi_times can be indexed together.
            >>>     for _imager, _imager_time in zip(asis.imagers, asi_times):
            >>>         info_str += f'{_imager.meta["location"]}: {_imager_time: %Y:%m:%d %H:%M:%S}\\n'
            >>>     info_str = info_str[:-1]  # Remove the training newline
            >>> 
            >>>     _text_obj = ax.text(
            >>>         0.01, 0.99, info_str, va='top', transform=ax.transAxes, 
            >>>         bbox=dict(facecolor='grey', edgecolor='black'))
            
        """
        if not overlap:
            self._calc_overlap_mask()

        if ax is None:
            ax = asilib.map.create_map(
                lon_bounds=lon_bounds,
                lat_bounds=lat_bounds,
                coast_color=coast_color,
                land_color=land_color,
                ocean_color=ocean_color,
            )

        # Create the animation directory inside asilib.config['ASI_DATA_DIR'] if it does
        # not exist.
        if animation_save_dir is None:
            _path = asilib.config['ASI_DATA_DIR']
        else:
            _path = animation_save_dir
        image_save_dir = pathlib.Path(
            _path,
            'animations',
            'images',
            f'{self.imagers[0].file_info["time_range"][0].strftime("%Y%m%d_%H%M%S")}_mosaic',
        )

        self.animation_name = (
            f'{self.imagers[0].file_info["time_range"][0].strftime("%Y%m%d_%H%M%S")}_'
            f'{self.imagers[0].file_info["time_range"][-1].strftime("%H%M%S")}_mosaic.'
            f'{movie_container}'
        )
        movie_save_path = image_save_dir.parents[1] / self.animation_name
        # If the image directory exists we need to first remove all of the images to avoid
        # animating images produced by different method calls.
        if image_save_dir.is_dir():
            shutil.rmtree(image_save_dir)
        image_save_dir.mkdir(parents=True)

        image_paths = []
        _progressbar = asilib.utils.progressbar(
            enumerate(self.__iter__()),
            iter_length=self.imagers[0]._estimate_n_times(),
            text=self.animation_name,
        )

        for i, (_guide_time, _asi_times, _asi_images) in _progressbar:
            asi_labels = len(self.imagers)*[None]
            pcolormesh_objs = len(self.imagers)*[None]
            for j, (_asi_time, _asi_image) in enumerate(zip(_asi_times, _asi_images)):
                if _asi_time == datetime.min:
                    continue
                _color_map, _color_norm = self.imagers[j]._plot_params(
                    _asi_image, color_bounds, color_map, color_norm
                    )

                ax, pcolormesh_objs[j], asi_labels[j] = self.imagers[j]._plot_mapped_image(
                    ax, _asi_image, min_elevation, _color_map, _color_norm, color_brighten, asi_label, 
                    pcolormesh_kwargs
                )

            # Give the user the control of the subplot, image object, and return the image time
            # so that they can manipulate the image to add, for example, the satellite track.
            yield _guide_time, _asi_times, _asi_images, ax

            # Save the plot before the next iteration.
            save_name = f'{str(i).zfill(6)}.png'
            plt.savefig(image_save_dir / save_name)
            image_paths.append(image_save_dir / save_name)

            # Clean up the objects that this method generated.
            for _asi_label in asi_labels:
                if _asi_label is not None:
                    _asi_label.remove()
            for pcolormesh_obj in pcolormesh_objs:
                if pcolormesh_obj is not None:
                    pcolormesh_obj.remove()
        
        self.imagers[0]._create_animation(image_paths, movie_save_path, ffmpeg_params, overwrite)
        return
    
    def __iter__(self) -> Generator[datetime, List, List]:
        """
        Generate a list of time stamps and images for all synchronized imagers, one time stamp 
        at a time.

        Yields
        ------
        datetime.datetime
            The guide time used to synchronize the imagers. If the difference between the imager
            time and the guide time is greater than time_tol*imager_cadence, or the imager is off,
            the imager is considered unsynchronized.
        list
            The time stamps from each imager. If the imager is unsynchronized the yielded 
            time stamp is ``datetime.min``.
        list 
            The images from each imager. If the imager is unsynchronized the yielded 
            image is ``None``.
        """
        t0 = self.imagers[0].file_info['time_range'][0]

        times = np.array(
            [t0+timedelta(seconds=i*self.imagers[0].meta['cadence']) 
            for i in range(self.imagers[0]._estimate_n_times())]
            )
        # asi_iterators keeps track of all ASIs in Imagers, with same order as passed into
        # __init__(). 
        # 
        # We must also keep track of ASIs whose next time stamp beyond the allowable tolerance 
        # of guide_time, or is off. future_iterator holds the imager (time, image) data until 
        # it becomes synchronized with the guide_time again.
        asi_iterators = {
            f'{_imager.meta["array"]}-{_imager.meta["location"]}':iter(_imager) 
            for _imager in self.imagers
            }
        future_iterators = {}
        stopped_iterators = []

        # TODO: Recalculate the skymaps if an imager is delayed or turned off.

        for guide_time in times:
            _asi_times = []
            _asi_images = []
            for _name, _iterator in asi_iterators.items():
                # We have three cases to address. If the iterator is synchronized, the
                # future_iterators will not have the _name key. This will trigger next() 
                # on that iterator. In either case, this will return a time stamp and 
                # an image. The one exception is when the _iterator is exhausted we fill 
                # in dummy values for the time and image.
                try:
                    _asi_time, _asi_image = future_iterators.get(_name, next(_iterator))
                except StopIteration:
                    _asi_times.append(datetime.min)
                    _asi_images.append(None)

                    if _name not in stopped_iterators:
                        stopped_iterators.append(_name)
                    if len(asi_iterators) == len(stopped_iterators):
                        # Stop once all iterators are exhausted.
                        return
                    continue
                abs_dt = np.abs((guide_time-_asi_time).total_seconds())
                synchronized = abs_dt < self.imagers[0].meta['cadence']*self.iter_tol

                # We must always append a time stamp and image, even if a dummy variable
                # to preserve the Imager order.
                if synchronized:
                    _asi_times.append(_asi_time)
                    _asi_images.append(_asi_image)
                else:
                    future_iterators[_name] = (_asi_time, _asi_image)
                    _asi_times.append(datetime.min)
                    _asi_images.append(None)

                if synchronized and (_name in future_iterators):
                    future_iterators.pop(_name)
            yield guide_time, _asi_times, _asi_images
        return

    def get_points(self, min_elevation:float=10)->Tuple[np.ndarray, np.ndarray]:
        """
        Get pixel intensities in each (lat, lon) grid point.

        Parameters
        ----------
        min_elevation: float
            Only return pixel intensities above min_elevation.

        Returns
        -------
        np.ndarray
            An (n, 2) array with each row corresponding to a (lat, lon) point.
        np.ndarray
            Pixel intensities with shape (n) for white-light images, and (n, 3) for RGB images.

        Examples
        --------
        >>> from datetime import datetime
        >>> 
        >>> import asilib
        >>> import asilib.asi
        >>> 
        >>> time = datetime(2007, 3, 13, 5, 8, 45)
        >>> location_codes = ['FSIM', 'ATHA', 'TPAS', 'SNKQ']
        >>> map_alt = 110
        >>> min_elevation = 2
        >>> 
        >>> _imagers = [asilib.asi.themis(location_code, time=time, alt=map_alt) 
        >>>             for location_code in location_codes]
        >>> asis = asilib.Imagers(_imagers)
        >>> lat_lon_points, intensities = asis.get_points(min_elevation=min_elevation)

        >>> # A comprehensive example showing how Imagers.get_points() can closely reproduce 
        >>> # Imagers.plot_map()
        >>> from datetime import datetime
        >>> 
        >>> import matplotlib.pyplot as plt
        >>> import matplotlib.colors
        >>> 
        >>> import asilib
        >>> import asilib.map
        >>> import asilib.asi
        >>> 
        >>> time = datetime(2007, 3, 13, 5, 8, 45)
        >>> location_codes = ['FSIM', 'ATHA', 'TPAS', 'SNKQ']
        >>> map_alt = 110
        >>> min_elevation = 2
        >>> 
        >>> _imagers = [asilib.asi.themis(location_code, time=time, alt=map_alt) 
        >>>             for location_code in location_codes]
        >>> asis = asilib.Imagers(_imagers)
        >>> lat_lon_points, intensities = asis.get_points(min_elevation=min_elevation)
        >>> 
        >>> fig = plt.figure(figsize=(12,5))
        >>> ax = asilib.map.create_simple_map(
        >>>     lon_bounds=(-140, -60), lat_bounds=(40, 82), fig_ax=(fig, 121)
        >>>     )
        >>> bx = asilib.map.create_simple_map(
        >>>     lon_bounds=(-140, -60), lat_bounds=(40, 82), fig_ax=(fig, 122)
        >>>     )
        >>> asis.plot_map(ax=ax, overlap=False, min_elevation=min_elevation)
        >>> bx.scatter(lat_lon_points[:, 1], lat_lon_points[:, 0], c=intensities, 
        >>>         norm=matplotlib.colors.LogNorm())
        >>> ax.text(0.01, 0.99, f'(A) Mosaic using Imagers.plot_map()', transform=ax.transAxes, 
        >>>         va='top', fontweight='bold', color='red')
        >>> bx.text(0.01, 0.99, f'(B) Mosaic from Imagers.get_points() scatter', transform=bx.transAxes,
        >>>         va='top', fontweight='bold', color='red')
        >>> fig.suptitle('Donovan et al. 2008 | First breakup of an auroral arc')
        >>> plt.tight_layout()
        >>> plt.show()
        """
        lat_lon_points = np.zeros((0, 2), dtype=float)
        if len(self.imagers[0].meta['resolution']) == 3: # RGB
            intensities  = np.zeros((0, self.imagers[0].meta['resolution'][-1]), dtype=float)
        else:  # single-color (or white light)
            intensities  = np.zeros(0, dtype=float)

        self._calc_overlap_mask()

        for _imager in self.imagers:
            assert 'time' in _imager.file_info.keys(), (
                f'Imagers.get_points() only works with single images.'
                )
            image = _imager.data.image
            _masked_lon_map, _masked_lat_map, _masked_image = _imager._mask_low_horizon(
                _imager.skymap['lon'], _imager.skymap['lat'], _imager.skymap['el'], min_elevation, 
                image=image
            )
            if _imager.skymap['lon'].shape[0] == image.shape[0] + 1:
                # Skymap defines vertices. We look for NaNs at either of the pixel edges.
                _valid_idx = np.where(
                    ~np.isnan(_masked_lon_map[1:, 1:]-_masked_lon_map[:-1, :-1])
                    )
            elif _imager.skymap['lon'].shape[0] == image.shape[0]:
                # Skymap defines pixel centers
                _valid_idx = np.where(~np.isnan(_masked_lon_map))
            else:
                raise ValueError(f'The skymap shape: {_imager.skymap["lon"].shape} and image '
                                 f'shape: {image.shape} are incompatible.')

            lat_grid = _masked_lat_map[_valid_idx[0], _valid_idx[1]]
            lon_grid = _masked_lon_map[_valid_idx[0], _valid_idx[1]]
            intensity = _masked_image[_valid_idx[0], _valid_idx[1], ...]

            # Concatenate joins arrays along an existing axis, while stack joins arrays
            # along a new axis. 
            intensities = np.concatenate((intensities, intensity))
            lat_lon_points = np.concatenate((
                lat_lon_points, 
                np.stack((lat_grid, lon_grid)).T
                ))
        return lat_lon_points, intensities
    
    def _calc_overlap_mask(self):
        """
        Calculate which pixels to plot for overlapping imagers by the criteria that the ith 
        imager's pixel must be closest to that imager (and not a neighboring one).

        Algorithm:
        1. Loop over ith imager
        2. Loop over jth imager within 500 km distance of the ith imager
        3. Mask low-elevations with np.nan
        4. Create a distance array with shape (resolution[0], resolution[1], j_total)
        5. For all pixels in ith imager, calculate the haversine distance to jth imager and 
        assign it to distance[..., j].
        6. For all pixels calculate the nearest imager out of all j's.
        7. If the minimum j is not the ith imager, mask the imager.skymap['lat'] and 
        imager.skymap['lon'] as np.nan.
        """
        if hasattr(self, '_masked'):
            return

        for i, imager in enumerate(self.imagers):
            _distances = np.nan*np.ones((*imager.skymap['lat'].shape, len(self.imagers)))
            for j, other_imager in enumerate(self.imagers):
                # Calculate the distance between all imager pixels and every other imager 
                # location (including itself).
                _distances[:, :, j] = _haversine(
                    imager.skymap['lat'], imager.skymap['lon'],
                    np.broadcast_to(other_imager.meta['lat'], imager.skymap['lat'].shape), 
                    np.broadcast_to(other_imager.meta['lon'], imager.skymap['lat'].shape)
                    )
            # Without this small reduction in the distance of pixels to its own imager,
            # there are gaps between the imager boundaries. In other words, this scaling
            # slightly biases to plotting pixels nearest to the imager. 
            _distances[:, :, i] *= 0.95
            # Need a masked array so that np.nanargmin correctly handles all NaN slices.
            _distances = np.ma.masked_array(_distances, np.isnan(_distances))
            # For each pixel, calculate the nearest imager. If the pixel is not closest to 
            # the imager that it's from, mask it as np.nan. Then the Imager._pcolormesh_nan() 
            # method then won't plot that pixel.
            min_distances = np.argmin(_distances, axis=2)
            far_pixels = np.where(min_distances != i)
            imager.skymap['lat'][far_pixels] = np.nan
            imager.skymap['lon'][far_pixels] = np.nan
        self._masked = True  # A flag to not run again.
        return
    
    def __str__(self):
        names = [f'{_img.meta["array"]}-{_img.meta["location"]}' for _img in self.imagers]
        names = 'asilib.Imagers initiated with:\n' + ', '.join(names)

        if (
                ('time' in self.imagers[0].file_info.keys()) and 
                (self.imagers[0].file_info['time'] is not None)
            ):
            return (names + f'\ntime={self.imagers[0].file_info["time"]}')
        elif (
                ('time_range' in self.imagers[0].file_info.keys()) and 
                (self.imagers[0].file_info['time_range'] is not None)
            ):
            return (names + f'\ntime_range={self.imagers[0].file_info["time_range"]}')
        else:
            raise ValueError(
                'The 0th imager object does not have a "time" or a "time_range" variable.'
                )
    

if __name__ == '__main__':
    # """
    # Animate a THEMIS ASI mosaic from Jones+2013 (https://doi.org/10.1002/jgra.50301)
    # """
    # import asilib
    # import asilib.asi

    # time_range = ('2008-02-11T04:22', '2008-02-11T04:45')
    # location_codes = [
    #     'INUV', 'FSIM', 'PGEO', 'FSMI', 'FSIM', 'ATHA', 'RANK', 'GILL', 'SNKQ', 'KUUJ'
    # ]
    # asi_list = []

    # for location_code in location_codes:
    #     asi_list.append(asilib.asi.themis(location_code, time_range=time_range))

    # asis = asilib.Imagers(asi_list)
    # gen = asis.animate_map_gen(overwrite=True)
    # for guide_time, asi_times, asi_images, ax in gen:
    #     if '_text_obj' in locals():
    #         _text_obj.remove()  # noqa: F821
    #     info_str = f'Time: {guide_time: %Y:%m:%d %H:%M:%S}'

    #     _text_obj = ax.text(
    #         0.01, 0.99, info_str, va='top', transform=ax.transAxes, 
    #         bbox=dict(facecolor='grey', edgecolor='black'))

    """
    Animate a REGO ASI mosaic from Panov+2019 (https://doi.org/10.1029/2019JA026521)
    """
    import matplotlib.pyplot as plt

    import asilib
    import asilib.asi
    import asilib.map

    time_range = ('2016-08-09T08:00', '2016-08-09T09:00')
    asi_list = []

    for location_code in ['GILL', 'FSMI', 'FSIM']:
        asi_list.append(asilib.asi.rego(location_code, time_range=time_range))

    ax = asilib.map.create_cartopy_map(lon_bounds=(-130, -87), lat_bounds=(51, 65))
    plt.tight_layout()

    asis = asilib.Imagers(asi_list)
    gen = asis.animate_map_gen(overwrite=True, ax=ax)
    for guide_time, asi_times, asi_images, ax in gen:
        if '_text_obj' in locals():
            _text_obj.remove()  # noqa: F821
        info_str = f'Time: {guide_time: %Y:%m:%d %H:%M:%S}'

        _text_obj = ax.text(
            0.01, 0.99, info_str, va='top', transform=ax.transAxes, 
            bbox=dict(facecolor='grey', edgecolor='black'))


import numpy as np
import matplotlib.colors as colors

# TODO: Make all of the other plot functions call these ones.


def get_color_bounds(image):
    """
    A decent default for the minimum and maximum colorbar values for aurora images. This way
    bright objects like the moon don't saturate the image while preserving enough dynamic range.
    """
    lower, upper = np.nanquantile(image, (0.25, 0.98))
    color_bounds = [lower, np.min([upper, lower * 10])]
    return color_bounds


def get_color_map(asi_array_code, color_map):
    """
    Color maps for the THEMIS and REGO ASIs.
    """
    if (color_map == 'auto') and (asi_array_code.lower() == 'themis'):
        color_map = 'Greys_r'
    elif (color_map == 'auto') and (asi_array_code.lower() == 'rego'):
        color_map = colors.LinearSegmentedColormap.from_list('black_to_red', ['k', 'r'])
    else:
        raise NotImplementedError('color_map == "auto" but the asi_array_code is unsupported')
    return color_map


def get_color_norm(color_norm, color_bounds):
    """
    Sets the normalization of the images to linear or logarithmic. You will typically call
    get_color_bounds() to get the color_bounds for this function.
    """
    if color_norm == 'log':
        norm = colors.LogNorm(vmin=color_bounds[0], vmax=color_bounds[1])
    elif color_norm == 'lin':
        norm = colors.Normalize(vmin=color_bounds[0], vmax=color_bounds[1])
    else:
        raise ValueError('color_norm must be either "log" or "lin".')
    return norm



import pathlib
import configparser

__version__ = '0.23.0'

# Load the configuration settings.
HERE = pathlib.Path(__file__).parent.resolve()
settings = configparser.ConfigParser()
settings.read(HERE / 'config.ini')

try:
    ASI_DATA_DIR = settings['Paths'].get('ASI_DATA_DIR', pathlib.Path.home() / 'asilib-data')
    ASI_DATA_DIR = pathlib.Path(ASI_DATA_DIR)
except KeyError:  # Raised if config.ini does not have Paths.
    ASI_DATA_DIR = pathlib.Path.home() / 'asilib-data'

config = {'ASILIB_DIR': HERE, 'ASI_DATA_DIR': ASI_DATA_DIR}

# Import download programs.
from asilib.io.download import download_image
from asilib.io.download import download_skymap

# Import the loading functions.
from asilib.io.load import load_skymap
from asilib.io.load import load_image
from asilib.io.load import load_image_generator

# Import the plotting and animating functions.
from asilib.plot.plot_fisheye import plot_fisheye
from asilib.plot.plot_map import plot_map
from asilib.plot.plot_map import make_map
from asilib.plot.plot_keogram import plot_keogram
from asilib.plot.animate_fisheye import animate_fisheye
from asilib.plot.animate_fisheye import animate_fisheye_generator
from asilib.plot.animate_map import animate_map
from asilib.plot.animate_map import animate_map_generator

# Import the analysis functions.
from asilib.analysis.map import lla2azel
from asilib.analysis.map import lla2footprint
from asilib.analysis.keogram import keogram
from asilib.analysis.equal_area import equal_area

# Imager implementation functions and classes.
from asilib.imager import Imager
from asilib.imagers import Imagers
from asilib.conjunction import Conjunction

__all__ = [
    'Imager',
    'Imagers',
    'Conjunction',
]  # So Sphinx shortens the name. See https://stackoverflow.com/a/31594545



import shutil
import importlib
import pathlib
from datetime import datetime
from typing import List, Union, Generator, Tuple
import warnings

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

import asilib
import asilib.plot.utils
from asilib.io.load import load_image
from asilib.io.load import load_skymap
from asilib.plot.plot_map import make_map
from asilib.plot.plot_map import _pcolormesh_nan
from asilib.analysis.start_generator import start_generator
from asilib.plot.animate_fisheye import _write_movie
from asilib.plot.animate_fisheye import Images


def animate_map(
    asi_array_code: str,
    location_code: str,
    time_range: asilib.io.utils._time_range_type,
    map_alt: float,
    **kwargs,
):
    """
    Projects a series of THEMIS or REGO images on a map at map_alt altitude in kilometers and
    animates them.

    This function basically runs ``animate_map_generator()`` in a for loop. The two function's
    arguments and keyword arguments are identical, so see ``animate_map_generator()`` docs for
    the full argument list.

    Note: To make animations, you'll need to install ``ffmpeg`` in your operating system.

    .. warning::
        Use :py:meth:`~asilib.imager.Imager.animate_map()` instead. This function will be
        removed in or after December 2023.

    Parameters
    ----------
    asi_array_code: str
        The imager array name, i.e. ``THEMIS`` or ``REGO``.
    location_code: str
        The ASI station code, i.e. ``ATHA``
    time_range: list of datetime.datetimes or stings
        Defined the duration of data to download. Must be of length 2.

    Returns
    -------
    None

    Raises
    ------
    NotImplementedError
        If the colormap is unspecified ('auto' by default) and the
        auto colormap is undefined for an ASI array.
    ValueError
        If the color_norm kwarg is not "log" or "lin".
    AssertionError
        If the ASI data exists for that time period, but without time stamps
        inside time_range.

    Example
    -------
    | from datetime import datetime
    |
    | import asilib
    |
    | map_alt=110  # km
    | time_range = (datetime(2015, 3, 26, 6, 7), datetime(2015, 3, 26, 6, 12))
    | asilib.animate_map('THEMIS', 'FSMI', time_range, map_alt=map_alt)
    | print(f'Movie saved in {asilib.config["ASI_DATA_DIR"] / "animations"}')
    """
    warnings.warn(
        "Use asilib.Imager.animate_map() instead. This function will be removed "
        "in or after December 2023.",
        DeprecationWarning,
    )

    map_generator = animate_map_generator(
        asi_array_code, location_code, time_range, map_alt, **kwargs
    )

    for _ in map_generator:
        pass
    return


@start_generator
def animate_map_generator(
    asi_array_code: str,
    location_code: str,
    time_range: asilib.io.utils._time_range_type,
    map_alt: float,
    min_elevation: float = 10,
    overwrite: bool = False,
    color_map: str = 'auto',
    color_bounds: Union[List[float], None] = None,
    color_norm: str = 'log',
    ax: plt.Axes = None,
    map_shapefile: Union[str, pathlib.Path] = 'ne_10m_land',
    coast_color: str = 'k',
    land_color: str = 'g',
    ocean_color: str = 'w',
    lon_bounds: tuple = (-140, -60),
    lat_bounds: tuple = (40, 82),
    label: bool = True,
    movie_container: str = 'mp4',
    ffmpeg_output_params={},
    pcolormesh_kwargs: dict = {},
) -> Generator[Tuple[datetime, np.ndarray, plt.Axes, matplotlib.image.AxesImage], None, None]:
    """
    Projects a series of THEMIS or REGO images on a map at map_alt altitude in kilometers and
    animates them. This generator function is useful if you need to superpose other data onto a
    map in the movie.

    Once this generator is initiated with the name `gen`, for example, but **before**
    the for loop, you can get the ASI images and times by calling `gen.send('data')`.
    This will yield a collections.namedtuple with `time` and `images` attributes.

    .. warning::
        Use :py:meth:`~asilib.imager.Imager.animate_map_gen()` instead. This function will be
        removed in or after December 2023.

    Parameters
    ----------
    asi_array_code: str
        The imager array name, i.e. ``THEMIS`` or ``REGO``.
    location_code: str
        The ASI station code, i.e. ``ATHA``
    time_range: list of datetime.datetimes or stings
        Defined the duration of data to download. Must be of length 2.
    map_alt: float
        The altitude in kilometers to project to. Must be an altitude value
        in the skymap calibration.
    min_elevation: float
        Masks the pixels below min_elevation degrees.
    overwrite: bool
        If True, the animation will be overwritten. Otherwise it will prompt
        the user to answer y/n.
    color_map: str
        The matplotlib colormap to use. If 'auto', will default to a
        black-red colormap for REGO and black-white colormap for THEMIS.
        For more information See
        https://matplotlib.org/3.3.3/tutorials/colors/colormaps.html
    color_bounds: List[float] or None
        The lower and upper values of the color scale. If None, will
        automatically set it to low=1st_quartile and
        high=min(3rd_quartile, 10*1st_quartile)
    ax: plt.Axes
        The optional subplot that will be drawn on.
    map_shapefile: str or pathlib.Path
        The path to the shapefile zip archive. If str, it will try to load the
        shapefile in asilib/data/{file}.
    coast_color: str
        The coast color. If None will not draw it.
    land_color: str
        The land color. If None will not draw it.
    ocean_color: str
        The ocean color. If None will not draw it.
    lon_bounds: tuple
        The map's longitude bounds.
    lat_bounds: tuple
        The map's latitude bounds.
    label: bool
        Annotates the map with the ASI code in the center of the image.
    movie_container: str
        The movie container: mp4 has better compression but avi was determined
        to be the official container for preserving digital video by the
        National Archives and Records Administration.
    ffmpeg_output_params: dict
        The additional/overwitten ffmpeg output parameters. The default parameters are:
        framerate=10, crf=25, vcodec=libx264, pix_fmt=yuv420p, preset=slower.
    color_norm: str
        Sets the 'lin' linear or 'log' logarithmic color normalization.
    pcolormesh_kwargs: dict
        A dictionary of keyword arguments (kwargs) to pass directly into
        plt.pcolormesh. One use of this parameter is to change the colormap. For example,
        pcolormesh_kwargs = {'cmap':'tu}

    Yields
    ------
    image_time: datetime.datetime
        The time of the current image.
    image: np.ndarray
        A 2d image array of the image corresponding to image_time
    ax: plt.Axes
        The subplot object to modify the axis, labels, etc.
    im: plt.AxesImage
        The plt.pcolormesh object.

    Raises
    ------
    NotImplementedError
        If the colormap is unspecified ('auto' by default) and the
        auto colormap is undefined for an ASI array.
    ValueError
        If the color_norm kwarg is not "log" or "lin".
    AssertionError
        If the ASI data exists for that time period, but without time stamps
        inside time_range.

    Example
    -------
    | from datetime import datetime
    |
    | import asilib
    |
    | map_alt=110
    | time_range = (datetime(2015, 3, 26, 6, 7), datetime(2015, 3, 26, 6, 12))
    | map_generator = asilib.animate_map_generator('THEMIS', 'FSMI', time_range, map_alt=map_alt,
    |     lon_bounds=(-125, -100), lat_bounds=(55, 70))
    |
    | for (image_time1, image, ax, p) in map_generator:
    |       # The code that modifies each image here.
    |       pass
    |
    | print(f'Movie saved in {asilib.config["ASI_DATA_DIR"] / "animations"}')
    """
    warnings.warn(
        "Use asilib.Imager.animate_fisheye_gen() instead. This function will be removed "
        "in or after December 2023.",
        DeprecationWarning,
    )
    try:
        image_times, images = load_image(asi_array_code, location_code, time_range=time_range)
    except AssertionError as err:
        if '0 number of time stamps were found in time_range' in str(err):
            print(
                f'The file exists for {asi_array_code}/{location_code}, but no data '
                f'between {time_range}.'
            )
            raise
        else:
            raise
    skymap = load_skymap(asi_array_code, location_code, time_range[0])

    # Create the movie directory inside asilib.config['ASI_DATA_DIR'] if it does
    # not exist.
    image_save_dir = pathlib.Path(
        asilib.config['ASI_DATA_DIR'],
        'animations',
        'images',
        f'{image_times[0].strftime("%Y%m%d_%H%M%S")}_{asi_array_code.lower()}_'
        f'{location_code.lower()}_map',
    )
    # If the image directory exists we need to first remove all of the images to avoid
    # animating images from different calls.
    if image_save_dir.is_dir():
        shutil.rmtree(image_save_dir)
    image_save_dir.mkdir(parents=True)
    print(f'Created a {image_save_dir} directory')

    # Check that the map_alt is in the skymap calibration data.
    assert (
        map_alt in skymap['FULL_MAP_ALTITUDE'] / 1000
    ), f'{map_alt} km is not in skymap calibration altitudes: {skymap["FULL_MAP_ALTITUDE"]/1000} km'
    alt_index = np.where(skymap['FULL_MAP_ALTITUDE'] / 1000 == map_alt)[0][0]
    image, lon_map, lat_map = _mask_low_horizon(
        images,
        skymap['FULL_MAP_LONGITUDE'][alt_index, :, :],
        skymap['FULL_MAP_LATITUDE'][alt_index, :, :],
        skymap['FULL_ELEVATION'],
        min_elevation,
    )

    if ax is None:
        ax = make_map(
            file=map_shapefile,
            coast_color=coast_color,
            land_color=land_color,
            ocean_color=ocean_color,
            lon_bounds=lon_bounds,
            lat_bounds=lat_bounds,
        )

    color_map = asilib.plot.utils.get_color_map(asi_array_code, color_map)

    # With the @start_generator decorator, when this generator first gets called, it
    # will halt here. This way the errors due to missing data will be raised up front.
    # user_input can be used to get the image_times and images out of the generator.
    user_input = yield
    if isinstance(user_input, str) and 'data' in user_input.lower():
        yield Images(image_times, images)

    if label:
        ax.text(
            skymap['SITE_MAP_LONGITUDE'],
            skymap['SITE_MAP_LATITUDE'],
            location_code.upper(),
            color='r',
            va='center',
            ha='center',
        )

    image_paths = []
    for i, (image_time, image) in enumerate(zip(image_times, images)):
        if 'p' in locals():
            p.remove()  # noqa

        # if-else statement is to recalculate color_bounds for every image
        # and set it to _color_bounds. If _color_bounds did not exist,
        # color_bounds will be overwritten after the first iteration which will
        # disable the dynamic color bounds for each image.
        if color_bounds is None:
            _color_bounds = asilib.plot.utils.get_color_bounds(image)
        else:
            _color_bounds = color_bounds

        norm = asilib.plot.utils.get_color_norm(color_norm, _color_bounds)

        p = _pcolormesh_nan(
            lon_map,
            lat_map,
            image,
            ax,
            cmap=color_map,
            norm=norm,
            pcolormesh_kwargs=pcolormesh_kwargs,
        )

        # Give the user the control of the subplot, image object, and return the image time
        # so that the user can manipulate the image to add, for example, the satellite track.
        yield image_time, image, ax, p

        # Save the plot before the next iteration.
        save_name = f'{str(i).zfill(5)}.png'
        plt.savefig(image_save_dir / save_name)
        image_paths.append(image_save_dir / save_name)

    # Make the movie
    movie_save_name = (
        f'{image_times[0].strftime("%Y%m%d_%H%M%S")}_'
        f'{image_times[-1].strftime("%H%M%S")}_'
        f'{asi_array_code.lower()}_{location_code.lower()}_map.{movie_container}'
    )
    movie_save_path = image_save_dir.parents[1] / movie_save_name
    _write_movie(image_paths, movie_save_path, ffmpeg_output_params, overwrite)
    return


def _mask_low_horizon(images, lon_map, lat_map, el_map, min_elevation):
    """
    Mask the images, skymap['FULL_MAP_LONGITUDE'], skymap['FULL_MAP_LONGITUDE'] arrays
    with np.nans where the skymap['FULL_ELEVATION'] is nan or
    skymap['FULL_ELEVATION'] < min_elevation.
    """
    idh = np.where(np.isnan(el_map) | (el_map < min_elevation))
    # Copy variables to not modify original np.arrays.
    images_copy = images.copy()
    lon_map_copy = lon_map.copy()
    lat_map_copy = lat_map.copy()
    # Can't mask image unless it is a float array.
    image_copy = images_copy.astype(float)
    image_copy[:, idh] = np.nan
    lon_map_copy[idh] = np.nan
    lat_map_copy[idh] = np.nan

    # For some reason the lat/lon_map arrays are one size larger than el_map, so
    # here we mask the boundary indices in el_map by adding 1 to both the rows
    # and columns.
    idh_boundary_bottom = (idh[0] + 1, idh[1])  # idh is a tuple so we have to create a new one.
    idh_boundary_right = (idh[0], idh[1] + 1)
    lon_map_copy[idh_boundary_bottom] = np.nan
    lat_map_copy[idh_boundary_bottom] = np.nan
    lon_map_copy[idh_boundary_right] = np.nan
    lat_map_copy[idh_boundary_right] = np.nan
    return images_copy, lon_map_copy, lat_map_copy



from __future__ import annotations  # to support the -> List[Downloader] return type
from typing import List
import pathlib
import urllib
import re

from bs4 import BeautifulSoup
import requests

import asilib
import asilib.io.utils as utils


themis_dir = pathlib.Path(asilib.config['ASI_DATA_DIR'], 'themis')
if not themis_dir.exists():
    themis_dir.mkdir(parents=True)

rego_dir = pathlib.Path(asilib.config['ASI_DATA_DIR'], 'rego')
if not rego_dir.exists():
    rego_dir.mkdir(parents=True)


def download_image(
    asi_array_code: str,
    location_code: str,
    time: utils._time_type = None,
    time_range: utils._time_range_type = None,
    redownload: bool = False,
    ignore_missing_data: bool = True,
) -> List[pathlib.Path]:
    """
    Download a CDF image file when ``time`` is given, or multiple files
    when ``time_range`` is given.

    Parameters
    ----------
    asi_array_code: str
        The imager array name, i.e. ``THEMIS`` or ``REGO``.
    location_code: str
        The ASI station code, i.e. ``ATHA``
    time: datetime.datetime or str
        The date and time to download of the data. If str, ``time`` must be in the
        ISO 8601 standard.
    time_range: list of datetime.datetimes or stings
        Defined the duration of data to download. Must be of length 2.
    redownload: bool
        If True, download the file even if it already exists. Useful if a prior
        data download was incomplete.
    ignore_missing_data: bool
        Flag to ignore the ``FileNotFoundError`` that is raised when ASI
        data is unavailable for that date-hour. Only useful when ``time_range``
        is passed.

    Returns
    -------
    download_paths: list
        A list of pathlib.Path objects that contain the downloaded file
        path(s).

    Example
    -------
    | from datetime import datetime
    |
    | import asilib
    |
    | asi_array_code = 'THEMIS'
    | location_code = 'LUCK'
    | time = datetime(2017, 4, 13, 5)
    | download_path = asilib.download_image(asi_array_code, location_code, time)
    """

    if asi_array_code.lower() == 'themis':
        paths = download_themis_img(
            location_code,
            time=time,
            time_range=time_range,
            redownload=redownload,
            ignore_missing_data=ignore_missing_data,
        )
    elif asi_array_code.lower() == 'rego':
        paths = download_rego_img(
            location_code,
            time=time,
            time_range=time_range,
            redownload=redownload,
            ignore_missing_data=ignore_missing_data,
        )
    return paths


def download_skymap(
    asi_array_code: str, location_code: str, redownload: bool = False
) -> List[pathlib.Path]:
    """
    Download all of the THEMIS or REGO skymap IDL .sav files.

    Parameters
    ----------
    asi_array_code: str
        The imager array name, i.e. ``THEMIS`` or ``REGO``.
    location_code: str
        The ASI station code, i.e. ``ATHA``
    redownload: bool
        If True, download the file even if it already exists. Useful if a prior
        data download was incomplete.

    Returns
    -------
    download_paths: list
        A list of pathlib.Path objects that contain the skymap file
        path(s).

    Example
    -------
    | import asilib
    |
    | asi_array_code = 'THEMIS'
    | location_code = 'LUCK'
    | asilib.download_skymap(asi_array_code, location_code)
    """
    if asi_array_code.lower() == 'themis':
        paths = download_themis_skymap(location_code, redownload=redownload)
    elif asi_array_code.lower() == 'rego':
        paths = download_rego_skymap(location_code, redownload=redownload)
    return paths


def download_themis_img(
    location_code: str,
    time: utils._time_type = None,
    time_range: utils._time_range_type = None,
    redownload: bool = False,
    ignore_missing_data: bool = True,
) -> List[pathlib.Path]:
    """
    Download one hourly THEMIS cdf file given the imager location and ``time``, or
    multiple hourly files given ``time_range``.

    Parameters
    ----------
    location_code: str
        The ASI station code, i.e. ``ATHA``
    time: datetime.datetime or str
        The date and time to download of the data. If str, ``time`` must be in the
        ISO 8601 standard.
    time_range: list of datetime.datetimes or stings
        Defined the duration of data to download. Must be of length 2.
    redownload: bool
        If True, download the file even if it already exists. Useful if a prior
        data download was incomplete.
    ignore_missing_data: bool
        Flag to ignore the FileNotFoundError that is raised when ASI
        data is unavailable for that date-hour. Only used when
        ``time_range`` is specified.

    Returns
    -------
    download_paths: list
        A list of pathlib.Path objects that contain the downloaded file
        path(s).

    Example
    -------
    from datetime import datetime

    import asilib

    | location_code = 'LUCK'
    | time = datetime(2017, 4, 13, 5)
    | asilib.download_themis_img(location_code, time)
    """
    base_url = 'http://themis.ssl.berkeley.edu/data/themis/thg/l1/asi'
    if (time is None) and (time_range is None):
        raise AttributeError('Neither time or time_range is specified.')
    elif (time is not None) and (time_range is not None):
        raise AttributeError('Both time and time_range can not be simultaneously specified.')

    elif time is not None:
        time = utils._validate_time(time)
        download_path = _download_one_img_file('themis', location_code, base_url, time, redownload)
        download_paths = [
            download_path
        ]  # List for constancy with the time_range code chunk output.

    elif time_range is not None:
        time_range = utils._validate_time_range(time_range)
        download_hours = utils._get_hours(time_range)
        download_paths = []

        for hour in download_hours:
            try:
                download_path = _download_one_img_file(
                    'themis', location_code, base_url, hour, redownload
                )
                download_paths.append(download_path)
            except NotADirectoryError:
                if ignore_missing_data:
                    continue
                else:
                    raise

    return download_paths


def download_themis_skymap(location_code: str, redownload: bool = False) -> List[pathlib.Path]:
    """
    Download all of the THEMIS skymap IDL .sav files.

    Parameters
    ----------
    location_code: str
        The ASI station code, i.e. ``ATHA``
    redownload: bool
        If True, download the file even if it already exists. Useful if a prior
        data download was incomplete.

    Returns
    -------
    None

    Example
    -------
    import asilib

    | location_code = 'LUCK'
    | asilib.download_themis_skymap(location_code)
    """
    # Create the skymap directory in data/themis/skymap/location_code
    save_dir = asilib.config['ASI_DATA_DIR'] / 'themis' / 'skymap' / location_code.lower()
    if not save_dir.is_dir():
        save_dir.mkdir(parents=True)
        print(f'Made directory at {save_dir}')

    base_url = 'https://data.phys.ucalgary.ca/sort_by_project/THEMIS/asi/skymaps'
    url = f'{base_url}/{location_code.lower()}/'

    # Look for all of the skymap hyperlinks, go in each one of them, and
    # download the .sav file.
    skymap_folders_relative = utils._search_hrefs(url, search_pattern=location_code.lower())
    download_paths = []

    for skymap_folder in skymap_folders_relative:
        skymap_folder_absolute = url + skymap_folder

        # Lastly, research for the skymap .sav file.
        skymap_name = utils._search_hrefs(skymap_folder_absolute, search_pattern='.sav')[0]
        skymap_save_name = skymap_name.replace('-%2B', '')  # Replace the unicode '+'.

        # Download if redownload=True or the file does not exist.
        download_path = pathlib.Path(save_dir, skymap_save_name)
        download_paths.append(download_path)
        if redownload or (not download_path.is_file()):
            utils._stream_large_file(skymap_folder_absolute + skymap_name, download_path)
    return download_paths


def download_rego_img(
    location_code: str,
    time: utils._time_type = None,
    time_range: utils._time_range_type = None,
    redownload: bool = False,
    ignore_missing_data: bool = True,
) -> List[pathlib.Path]:
    """
    Download one hourly REGO cdf file given the imager location and ``time``, or
    multiple hourly files given ``time_range``.

    Parameters
    ----------
    location_code: str
        The ASI station code, i.e. ``ATHA``
    time: datetime.datetime or str
        The date and time to download of the data. If str, ``time`` must be in the
        ISO 8601 standard.
    time_range: list of datetime.datetimes or stings
        Defined the duration of data to download. Must be of length 2.
    redownload: bool
        If True, download the file even if it already exists. Useful if a prior
        data download was incomplete.
    ignore_missing_data: bool
        Flag to ignore the FileNotFoundError that is raised when ASI
        data is unavailable for that date-hour. Only used when
        ``time_range`` is specified.

    Returns
    -------
    download_paths: list
        A list of pathlib.Path objects that contain the downloaded file
        path(s).

    Example
    -------
    | from datetime import datetime
    |
    | import asilib
    |
    | location_code = 'LUCK'
    | time = datetime(2017, 4, 13, 5)
    | asilib.download_rego_img(location_code, time=time)
    """
    base_url = 'http://themis.ssl.berkeley.edu/data/themis/thg/l1/reg'

    if (time is None) and (time_range is None):
        raise AttributeError('Neither time or time_range is specified.')
    elif (time is not None) and (time_range is not None):
        raise AttributeError('Both time and time_range can not be simultaneously specified.')

    elif time is not None:
        time = utils._validate_time(time)
        download_path = _download_one_img_file('rego', location_code, base_url, time, redownload)
        download_paths = [
            download_path
        ]  # List for constancy with the time_range code chunk output.

    elif time_range is not None:
        time_range = utils._validate_time_range(time_range)
        download_hours = utils._get_hours(time_range)
        download_paths = []

        for hour in download_hours:
            try:
                download_path = _download_one_img_file(
                    'rego', location_code, base_url, hour, redownload
                )
                download_paths.append(download_path)
            except NotADirectoryError:
                if ignore_missing_data:
                    continue
                else:
                    raise

    return download_paths


def download_rego_skymap(location_code: str, redownload: bool = False) -> List[pathlib.Path]:
    """
    Download all of the REGO skymap IDL .sav files.

    Parameters
    ----------
    location_code: str
        The ASI station code, i.e. ``ATHA``
    redownload: bool
        If True, download the file even if it already exists. Useful if a prior
        data download was incomplete.

    Returns
    -------
    None

    Example
    -------
    | import asilib
    |
    | location_code = 'LUCK'
    | asilib.download_rego_skymap(location_code)
    """
    base_url = 'https://data.phys.ucalgary.ca/sort_by_project/GO-Canada/REGO/skymap'

    save_dir = asilib.config['ASI_DATA_DIR'] / 'rego' / 'skymap' / location_code.lower()
    if not save_dir.is_dir():
        save_dir.mkdir(parents=True)
        print(f'Made directory at {save_dir}')

    url = f'{base_url}/{location_code.lower()}/'

    # Look for all of the skymap hyperlinks, go in each one of them, and
    # download the .sav file.
    skymap_folders_relative = utils._search_hrefs(url, search_pattern=location_code.lower())
    download_paths = []

    for skymap_folder in skymap_folders_relative:
        skymap_folder_absolute = url + skymap_folder

        # Lastly, research for the skymap .sav file.
        skymap_name = utils._search_hrefs(skymap_folder_absolute, search_pattern='.sav')[0]
        skymap_save_name = skymap_name.replace('-%2B', '')  # Replace the unicode '+'.

        # Download if redownload=True or the file does not exist.
        download_path = pathlib.Path(save_dir, skymap_save_name)
        download_paths.append(download_path)
        if redownload or (not download_path.is_file()):
            utils._stream_large_file(skymap_folder_absolute + skymap_name, download_path)
    return download_paths


def _download_one_img_file(asi_array_code, location_code, base_url, time, redownload):
    """
    Download one hour-long file.
    """
    # Add the location/year/month url folders onto the url
    url = f'{base_url}/{location_code.lower()}/{time.year}/{str(time.month).zfill(2)}/'

    search_pattern = f'{location_code.lower()}_{time.strftime("%Y%m%d%H")}'
    file_names = utils._search_hrefs(url, search_pattern=search_pattern)

    server_url = url + file_names[0]
    download_path = pathlib.Path(
        asilib.config['ASI_DATA_DIR'], asi_array_code.lower(), file_names[0]
    )
    if redownload or (not download_path.is_file()):
        utils._stream_large_file(server_url, download_path)
    return download_path


# TODO: When cleaning up Imager, remove all above
class Downloader:
    """
    Traverses and lists the directory structure on a server and download files.

    Parameters
    ----------
    url: str
        The dataset URL. If it points to a folder it must end with a "/".
    download_dir: str or pathlib.Path
        The download directory. Must either specify here, or when you call
        Downloader.download().

    Example
    -------
    TODO: Add an example.
    """

    def __init__(self, url: str, download_dir=None) -> None:
        self.url = url
        self.download_dir = download_dir
        return

    def ls(self, match: str = '*') -> List[Downloader]:
        """
        List files and folders in self.url.

        Parameters
        ----------
        match: str
            An optional string pattern to match.

        Return
        ------
        list
            A list of full URLs.
        """
        matched_hrefs = self._search_hrefs(self.url, match=match)
        cls = type(self)
        downloaders = [None] * len(matched_hrefs)
        for i, matched_href in enumerate(matched_hrefs):
            new_url = urllib.parse.urljoin(self.url, matched_href, allow_fragments=True)
            downloaders[i] = cls(new_url, download_dir=self.download_dir)
        return downloaders

    def download(
        self, download_dir=None, redownload: bool = False, stream: bool = False
    ) -> pathlib.Path:
        """
        Downloads file from self.url to the download_dir directory.

        Parameters
        ----------
        download_dir: str or pathlib.Path
            The parent directory where to save the data to. Set it either
            here or when initializing the class. This
        redownload: bool
            Will redownload an existing file.
        stream: bool
            Download the data in one chunk if False, or break it up and download it
            in 5 Mb chunks if True.

        Returns
        -------
        pathlib.Path
            The full path to the file.
        """
        if download_dir is None and self.download_dir is None:
            raise ValueError(
                f'download_dir kwarg needs to be set either '
                f'in Downloader() or Downloader.download.'
            )
        if download_dir is not None:
            self.download_dir = download_dir

        self.download_dir = pathlib.Path(self.download_dir)
        self.download_dir.mkdir(parents=True, exist_ok=True)
        file_name = pathlib.Path(self.url).name
        download_path = self.download_dir / file_name

        if (download_path.exists()) and (not redownload):
            return download_path

        if stream:
            with requests.Session() as s:
                r = s.get(self.url, stream=True, timeout=5)
            r.raise_for_status()
            file_size = int(r.headers.get('content-length'))
            downloaded_bites = 0

            megabyte = 1024 * 1024
            try:
                with open(download_path, 'wb') as f:
                    for data in r.iter_content(chunk_size=10 * megabyte):
                        f.write(data)
                        # Update the downloaded % in the terminal.
                        downloaded_bites += len(data)
                        download_percent = round(100 * downloaded_bites / file_size)
                        download_str = "#" * (download_percent // 5)
                        print(
                            f'Downloading {file_name}: |{download_str:<20}| {download_percent}%',
                            end='\r',
                        )
                print()  # Add a newline
            except (Exception, KeyboardInterrupt, SystemExit) as err:
                download_path.unlink()
                raise RuntimeError(
                    f'Download interrupted. Partially downloaded file ' f'{download_path} deleted.'
                ) from err
        else:
            r = requests.get(self.url, timeout=5)
            with open(download_path, 'wb') as f:
                f.write(r.content)
            print(f'Downloaded {file_name}.')
        return download_path

    @property
    def name(self):
        """
        Get the url filename
        """
        _url = pathlib.Path(self.url)
        if _url.suffix != '':
            # url points to a filename. pathlib is not designed for
            # urls, but this is the easiest way to get the name.
            return _url.name
        else:
            return None

    def _search_hrefs(self, url: str, match: str = '*') -> List[str]:
        """
        Given a url string, find all hyper references matching the
        string match. The re module does the matching.

        Parameters
        ----------
        url: str
            A url
        match: str (optional)
            The regex match to compare the hyper references to. The default is
            to match everything (a wildcard)

        Returns
        -------
        List(str)
            A list of hyper references that matched the match string.

        Raises
        ------
        FileNotFoundError
            If no hyper references were found.
        """
        with requests.Session() as s:
            request = s.get(url, timeout=5)
        request.raise_for_status()
        soup = BeautifulSoup(request.content, 'html.parser')

        match = match.replace('*', '.*')  # regex wildcard
        href_objs = soup.find_all('a', href=re.compile(match))
        # I found "?" to be in the column names so we should ignore them.
        matched_hrefs = [
            href_obj['href'] for href_obj in href_objs if href_obj['href'].find('?') == -1
        ]
        if len(matched_hrefs) == 0:
            raise FileNotFoundError(
                f'The url {url} does not contain any hyper '
                f'references containing the match kwarg="{match}".'
            )
        return matched_hrefs

    def __repr__(self) -> str:
        params = f'{self.url}, download_dir={self.download_dir},'
        return f'{self.__class__.__qualname__}(' + params + ')'

    def __str__(self) -> str:
        return (
            f'{self.__class__.__qualname__} with url={self.url} and '
            f'download_dir={self.download_dir}'
        )



import pathlib
import shutil
from typing import List, Union, Generator, Tuple
from datetime import datetime
import collections
import warnings

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import numpy as np
import ffmpeg

import asilib
from asilib.io import utils
from asilib.io.load import load_image, load_skymap
from asilib.analysis.start_generator import start_generator


def animate_fisheye(
    asi_array_code: str, location_code: str, time_range: utils._time_range_type, **kwargs
) -> None:
    """
    Animate a series of THEMIS or REGO fisheye images.

    This function basically runs ``animate_fisheye_generator()`` in a for loop. The two function's
    arguments and keyword arguments are identical, so see ``animate_fisheye_generator()`` docs for
    the full argument list.

    Note: To make movies, you'll need to install ``ffmpeg`` in your operating system.

    .. warning::
        Use :py:meth:`~asilib.imager.Imager.animate_fisheye()` instead. This function will be
        removed in or after December 2023.

    Parameters
    ----------
    asi_array_code: str
        The imager array name, i.e. ``THEMIS`` or ``REGO``.
    location_code: str
        The ASI station code, i.e. ``ATHA``
    time_range: list of datetime.datetimes or stings
        Defined the duration of data to download. Must be of length 2.

    Returns
    -------
    None

    Raises
    ------
    NotImplementedError
        If the colormap is unspecified ('auto' by default) and the
        auto colormap is undefined for an ASI array.
    ValueError
        If the color_norm kwarg is not "log" or "lin".

    Example
    -------
    | from datetime import datetime
    |
    | import asilib
    |
    | time_range = (datetime(2015, 3, 26, 6, 7), datetime(2015, 3, 26, 6, 12))
    | asilib.animate_fisheye('THEMIS', 'FSMI', time_range)
    | print(f'Movie saved in {asilib.config["ASI_DATA_DIR"] / "movies"}')
    """

    # Create a subplot object if one is not passed.
    ax = kwargs.get('ax', None)
    if ax is None:
        _, ax = plt.subplots(figsize=(6, 6))
        kwargs['ax'] = ax
        plt.tight_layout()

    movie_generator = animate_fisheye_generator(asi_array_code, location_code, time_range, **kwargs)

    for image_time, image, im, ax in movie_generator:
        pass
    return


Images = collections.namedtuple('Images', ['time', 'images'])


@start_generator
def animate_fisheye_generator(
    asi_array_code: str,
    location_code: str,
    time_range: utils._time_range_type,
    overwrite: bool = False,
    label: bool = True,
    color_map: str = 'auto',
    color_bounds: Union[List[float], None] = None,
    color_norm: str = 'log',
    azel_contours: bool = False,
    ax: plt.Axes = None,
    movie_container: str = 'mp4',
    ffmpeg_output_params={},
) -> Generator[Tuple[datetime, np.ndarray, plt.Axes, matplotlib.image.AxesImage], None, None]:
    """
    A generator function that loads the ASI data and then yields individual ASI images,
    image by image. This allows the user to add content to each image, such as the
    spacecraft position, and that will convert it to a movie. If you just want to make
    an ASI fisheye movie, use the wrapper for this function, called animate_fisheye().

    Once this generator is initiated with the name `gen`, for example, but **before**
    the for loop, you can get the ASI images and times by calling `gen.send('data')`.
    This will yield a collections.namedtuple with `time` and `images` attributes.

    .. warning::
        Use :py:meth:`~asilib.imager.Imager.animate_fisheye_gen()` instead. This function will be
        removed in or after December 2023.

    Parameters
    ----------
    asi_array_code: str
        The imager array name, i.e. ``THEMIS`` or ``REGO``.
    location_code: str
        The ASI station code, i.e. ``ATHA``
    time_range: list of datetime.datetimes or stings
        Defined the duration of data to download. Must be of length 2.
    overwrite: bool
        If true, the output animation will be overwritten, otherwise it will
        prompt the user to answer y/n.
    label: bool
        Flag to add the "asi_array_code/location_code/image_time" text to the plot.
    color_map: str
        The matplotlib colormap to use. If 'auto', will default to a
        black-red colormap for REGO and black-white colormap for THEMIS.
        For more information See
        https://matplotlib.org/3.3.3/tutorials/colors/colormaps.html
    color_bounds: List[float] or None
        The lower and upper values of the color scale. If None, will
        automatically set it to low=1st_quartile and
        high=min(3rd_quartile, 10*1st_quartile)
    ax: plt.Axes
        The optional subplot that will be drawn on.
    movie_container: str
        The movie container: mp4 has better compression but avi was determined
        to be the official container for preserving digital video by the
        National Archives and Records Administration.
    ffmpeg_output_params: dict
        The additional/overwitten ffmpeg output parameters. The default parameters are:
        framerate=10, crf=25, vcodec=libx264, pix_fmt=yuv420p, preset=slower.
    color_norm: str
        Sets the 'lin' linear or 'log' logarithmic color normalization.
    azel_contours: bool
        Switch azimuth and elevation contours on or off.

    Yields
    ------
    image_time: datetime.datetime
        The time of the current image.
    image: np.ndarray
        A 2d image array of the image corresponding to image_time
    ax: plt.Axes
        The subplot object to modify the axis, labels, etc.
    im: plt.AxesImage
        The plt.imshow image object. Common use for im is to add a colorbar.
        The image is oriented in the map orientation (north is up, south is down,
        west is right, and east is left). Set azel_contours=True to confirm.

    Raises
    ------
    NotImplementedError
        If the colormap is unspecified ('auto' by default) and the
        auto colormap is undefined for an ASI array.
    ValueError
        If the color_norm kwarg is not "log" or "lin".

    Example
    -------
    | from datetime import datetime
    |
    | import asilib
    |
    | time_range = (datetime(2015, 3, 26, 6, 7), datetime(2015, 3, 26, 6, 12))
    | movie_generator = asilib.animate_fisheye_generator('THEMIS', 'FSMI', time_range)
    |
    | for image_time, image, im, ax in movie_generator:
    |       # The code that modifies each image here.
    |       pass
    |
    | print(f'Movie saved in {asilib.config["ASI_DATA_DIR"] / "movies"}')
    """
    warnings.warn(
        "Use asilib.Imager.animate_fisheye_gen() instead. This function will be removed in "
        "or after December 2023.",
        DeprecationWarning,
    )
    try:
        image_times, images = load_image(asi_array_code, location_code, time_range=time_range)
    except AssertionError as err:
        if '0 number of time stamps were found in time_range' in str(err):
            print(
                f'The file exists for {asi_array_code}/{location_code}, but no data '
                f'between {time_range}.'
            )
            raise
        else:
            raise
    if ax is None:
        _, ax = plt.subplots()

    # Create the movie directory inside asilib.config['ASI_DATA_DIR'] if it does
    # not exist.
    image_save_dir = pathlib.Path(
        asilib.config['ASI_DATA_DIR'],
        'animations',
        'images',
        f'{image_times[0].strftime("%Y%m%d_%H%M%S")}_{asi_array_code.lower()}_'
        f'{location_code.lower()}_fisheye',
    )
    # If the image directory exists we need to first remove all of the images to avoid
    # animating images from different calls.
    if image_save_dir.is_dir():
        shutil.rmtree(image_save_dir)
    image_save_dir.mkdir(parents=True)
    print(f'Created a {image_save_dir} directory')

    if (color_map == 'auto') and (asi_array_code.lower() == 'themis'):
        color_map = 'Greys_r'
    elif (color_map == 'auto') and (asi_array_code.lower() == 'rego'):
        color_map = colors.LinearSegmentedColormap.from_list('black_to_red', ['k', 'r'])
    else:
        raise NotImplementedError('color_map == "auto" but the asi_array_code is unsupported')

    # With the @start_generator decorator, when this generator first gets called, it
    # will halt here. This way the errors due to missing data will be raised up front.
    user_input = yield
    # user_input can be used to get the image_times and images out of the generator.
    if isinstance(user_input, str) and 'data' in user_input.lower():
        yield Images(image_times, images)

    image_paths = []
    for i, (image_time, image) in enumerate(zip(image_times, images)):
        # If the image is all 0s we have a bad image and we need to skip it.
        if np.all(image == 0):
            continue
        ax.clear()
        ax.axis('off')
        # if-else statement is to recalculate color_bounds for every image
        # and set it to _color_bounds. If _color_bounds did not exist,
        # color_bounds will be overwritten after the first iteration which will
        # disable the dynamic color bounds for each image.
        if color_bounds is None:
            lower, upper = np.quantile(image, (0.25, 0.98))
            _color_bounds = [lower, np.min([upper, lower * 10])]
        else:
            _color_bounds = color_bounds

        if color_norm == 'log':
            norm = colors.LogNorm(vmin=_color_bounds[0], vmax=_color_bounds[1])
        elif color_norm == 'lin':
            norm = colors.Normalize(vmin=_color_bounds[0], vmax=_color_bounds[1])
        else:
            raise ValueError('color_norm must be either "log" or "lin".')

        im = ax.imshow(image, cmap=color_map, norm=norm, origin='lower')
        if label:
            ax.text(
                0,
                0,
                f"{asi_array_code.upper()}/{location_code.upper()}\n{image_time.strftime('%Y-%m-%d %H:%M:%S')}",
                va='bottom',
                transform=ax.transAxes,
                color='white',
            )

        if azel_contours:
            _add_azel_contours(asi_array_code, location_code, image_time, ax)

        # Give the user the control of the subplot, image object, and return the image time
        # so that the user can manipulate the image to add, for example, the satellite track.
        yield image_time, image, ax, im

        # Save the plot before the next iteration.
        save_name = f'{str(i).zfill(5)}.png'
        plt.savefig(image_save_dir / save_name)
        image_paths.append(image_save_dir / save_name)

    # Make the movie
    movie_save_name = (
        f'{image_times[0].strftime("%Y%m%d_%H%M%S")}_'
        f'{image_times[-1].strftime("%H%M%S")}_'
        f'{asi_array_code.lower()}_{location_code.lower()}_fisheye.{movie_container}'
    )
    movie_save_path = image_save_dir.parents[1] / movie_save_name
    _write_movie(image_paths, movie_save_path, ffmpeg_output_params, overwrite)
    return


def _write_movie(image_paths, movie_save_path, ffmpeg_output_params, overwrite):
    """
    Helper function to write a movie using ffmpeg.

    Parameters
    ----------
    image_save_dir: pathlib.Path
        The directory where the individual images are saved to.
    ffmpeg_output_params: dict
        The additional/overwitten ffmpeg output parameters. The default parameters are:
        framerate=10, crf=25, vcodec=libx264, pix_fmt=yuv420p, preset=slower.
    movie_file_name: str
        The movie file name.
    overwrite: bool
        Overwrite the movie.
    """
    ffmpeg_params = {
        'framerate': 10,
        'crf': 25,
        'vcodec': 'libx264',
        'pix_fmt': 'yuv420p',
        'preset': 'slower',
    }
    # Add or change the ffmpeg_params's key:values with ffmpeg_output_params
    ffmpeg_params.update(ffmpeg_output_params)

    try:
        movie_obj = ffmpeg.input(
            str(image_paths[0].parent / "%05d.png"),
            pattern_type='sequence',
            # Pop so it won't be passed into movie_obj.output().
            framerate=ffmpeg_params.pop('framerate'),
        )
        movie_obj.output(str(movie_save_path), **ffmpeg_params).run(overwrite_output=overwrite)
    except FileNotFoundError as err:
        if '[WinError 2] The system cannot find the file specified' in str(err):
            raise FileNotFoundError("Windows doesn't have ffmpeg installed.") from err
    return


def _add_azel_contours(
    asi_array_code: str,
    location_code: str,
    time: utils._time_type,
    ax: plt.Axes,
    color: str = 'yellow',
) -> None:
    """
    Adds contours of azimuth and elevation to the movie image.

    Parameters
    ----------
    asi_array_code: str
        The asi_array_code, can be either THEMIS or REGO.
    location_code: str
        The imager location code to download the data from.
    time: datetime, or str
        Time is used to find the relevant skymap file: file created nearest to, and before, the time.
    ax: plt.Axes
        The subplot that will be drawn on.
    color: str (optional)
        The contour color.
    """
    skymap_dict = load_skymap(asi_array_code, location_code, time)

    az_contours = ax.contour(
        skymap_dict['FULL_AZIMUTH'],
        colors=color,
        linestyles='dotted',
        levels=np.arange(0, 361, 90),
        alpha=1,
    )
    el_contours = ax.contour(
        skymap_dict['FULL_ELEVATION'],
        colors=color,
        linestyles='dotted',
        levels=np.arange(0, 91, 30),
        alpha=1,
    )
    plt.clabel(az_contours, inline=True, fontsize=12, colors=color)
    plt.clabel(el_contours, inline=True, fontsize=12, colors=color, rightside_up=True)
    return



from typing import Tuple
import importlib

import numpy as np
import pymap3d
import scipy.spatial

try:
    import IRBEM
except ImportError:
    pass  # make sure that asilb.__init__ fully loads and crashes if the user calls asilib.lla2footprint().

import asilib
from asilib.io import utils


def lla2azel(
    asi_array_code: str, location_code: str, time: utils._time_type, sat_lla: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Maps, a satellite's latitude, longitude, and altitude (LLA) coordinates
    to the ASI's azimuth and elevation (azel) coordinates and image pixel index.

    This function is useful to plot a satellite's location in the ASI image using the
    pixel indices.

    Parameters
    ----------
    asi_array_code: str
        The imager array name, i.e. ``THEMIS`` or ``REGO``.
    location_code: str
        The ASI station code, i.e. ``ATHA``
    time: datetime.datetime or str
        The date and time to find the relevant skymap file. If str, ``time``
        must be in the ISO 8601 standard.
    sat_lla: np.ndarray or pd.DataFrame
        The satellite's latitude, longitude, and altitude coordinates in a 2d array
        with shape (nPosition, 3) where each row is the number of satellite positions
        to map, and the columns correspond to latitude, longitude, and altitude,
        respectively. The altitude is in kilometer units.

    Returns
    -------
    sat_azel: np.ndarray
        An array with shape (nPosition, 2) of the satellite's azimuth and
        elevation coordinates.
    asi_pixels: np.ndarray
        An array with shape (nPosition, 2) of the x- and y-axis pixel
        indices for the ASI image.

    Raises
    ------
    AssertionError
        If the sat_lla argument does not have exactly 3 columns (1st dimension).

    Example
    -------
    | from datetime import datetime
    |
    | import numpy as np
    | from asilib import lla2azel
    |
    | # THEMIS/ATHA's LLA coordinates are (54.72, -113.301, 676 (meters)).
    | # The LLA is a North-South pass right above ATHA..
    | n = 50
    | lats = np.linspace(60, 50, n)
    | lons = -113.64*np.ones(n)
    | alts = 500**np.ones(n)
    | lla = np.array([lats, lons, alts]).T
    |
    | time = datetime(2015, 10, 1)  # To load the proper skymap file.
    |
    | azel, pixels = lla2azel('REGO', 'ATHA', time, lla)
    """

    # Check the sat_lla input parameter to make sure it is of the correct shape
    input_shape = sat_lla.shape
    # Reshape into a (1,3) array if the user provided only one set of satellite
    # coordinates.
    if len(input_shape) == 1:
        sat_lla = sat_lla.reshape(1, input_shape[0])
    assert sat_lla.shape[1] == 3, 'sat_lla must have 3 columns.'

    # Load the catalog
    skymap_dict = asilib.io.load.load_skymap(asi_array_code, location_code, time)

    sat_azel = np.nan * np.zeros((sat_lla.shape[0], 2))

    # Loop over every set of LLA coordinates and use pymap3d.geodetic2aer
    # to map to the azimuth and elevation.
    for i, (lat_i, lon_i, alt_km_i) in enumerate(sat_lla):
        # Check if lat, lon, or alt is nan or -1E31
        # (the error value in the IRBEM library).
        any_nan = bool(len(np.where(np.isnan([lat_i, lon_i, alt_km_i]))[0]))
        any_neg = bool(len(np.where(np.array([lat_i, lon_i, alt_km_i]) == -1e31)[0]))
        if any_nan or any_neg:
            continue

        az, el, _ = pymap3d.geodetic2aer(
            lat_i,
            lon_i,
            1e3 * alt_km_i,
            skymap_dict['SITE_MAP_LATITUDE'],
            skymap_dict['SITE_MAP_LONGITUDE'],
            skymap_dict['SITE_MAP_ALTITUDE'],
        )
        sat_azel[i, :] = [az, el]

    # Now find the corresponding x- and y-axis pixel indices.
    asi_pixels = _map_azel_to_pixel(sat_azel, skymap_dict)

    # If len(input_shape) == 1, a 1d array, flatten the (1x3) sat_azel and
    # asi_pizels arrays into a (3,) array. This way the input and output
    # lla arrays have the same number of dimentions.
    if len(input_shape) == 1:
        return sat_azel.flatten(), asi_pixels.flatten()
    else:
        return sat_azel, asi_pixels


def lla2footprint(
    space_time: np.ndarray,
    map_alt: float,
    b_model: str = 'OPQ77',
    maginput: dict = None,
    hemisphere: int = 0,
) -> np.ndarray:
    """
    Map the spacecraft's position to ``map_alt`` along the magnetic field line.
    The mapping is implemeneted in ``IRBEM`` and by default it maps to the same
    hemisphere.

    Parameters
    ----------
    space_time: np.ndarray
        A 2d array with shape (n_times, 4) with the columns containing the
        time, latitude, longitude, and altitude coordinates in that order.
    map_alt: float
        The altitude to map to, in km, in the same hemisphere.
    b_model: str
        The magnetic field model to use, by default the model is Olson-Pfitzer
        1974. This parameter is passed directly into IRBEM.MagFields as the
        'kext' parameter.
    maginput: dict
        If you use a differnet b_model that requires time-dependent parameters,
        supply the appropriate values to the maginput dictionary. It is directly
        passed into IRBEM.MagFields so refer to IRBEM on the proper format.
    hemisphere: int
        The hemisphere to map to. This kwarg is passed to IRBEM and can be one of
        these four values:
        0    = same magnetic hemisphere as starting point
        +1   = northern magnetic hemisphere
        -1   = southern magnetic hemisphere
        +2   = opposite magnetic hemisphere as starting point

    Returns
    -------
    magnetic_footprint: np.ndarray
        A numpy.array with size (n_times, 3) with lat, lon, alt
        columns representing the magnetic footprint coordinates.

    Raises
    ------
    ImportError
        If IRBEM can't be imported.
    """
    if importlib.util.find_spec('IRBEM') is None:
        raise ImportError(
            "IRBEM can't be imported. This is a required dependency for asilib.lla2footprint()"
            " that must be installed separately. See https://github.com/PRBEM/IRBEM"
            " and https://aurora-asi-lib.readthedocs.io/en/latest/installation.html."
        )

    magnetic_footprint = np.nan * np.zeros((space_time.shape[0], 3))

    m = IRBEM.MagFields(kext=b_model)  # Initialize the IRBEM model.

    # Loop over every set of satellite coordinates.
    for i, (time, lat, lon, alt) in enumerate(space_time):
        X = {'datetime': time, 'x1': alt, 'x2': lat, 'x3': lon}
        m_output = m.find_foot_point(X, maginput, map_alt, hemisphere)
        magnetic_footprint[i, :] = m_output['XFOOT']
    # Map from IRBEM's (alt, lat, lon) -> (lat, lon, alt)
    magnetic_footprint[:, [2, 0, 1]] = magnetic_footprint[:, [0, 1, 2]]
    return magnetic_footprint


def _map_azel_to_pixel(sat_azel: np.ndarray, skymap_dict: dict) -> np.ndarray:
    """
    Given the 2d array of the satellite's azimuth and elevation, locate
    the nearest ASI skymap x- and y-axis pixel indices. Note that the
    scipy.spatial.KDTree() algorithm is efficient at finding nearest
    neighbors. However it does not work in a polar coordinate system.

    Parameters
    ----------
    sat_azel : array
        A 1d or 2d array of satelite azimuth and elevation points.
        If 2d, the rows correspons to time.
    skymap_dict: dict
        The skymap file dictionary

    Returns
    -------
    pixel_index: np.ndarray
        An array with the same shape as sat_azel, but representing the
        x- and y-axis pixel indices in the ASI image.
    """
    az_coords = skymap_dict['FULL_AZIMUTH'].ravel()
    az_coords[np.isnan(az_coords)] = -10000
    el_coords = skymap_dict['FULL_ELEVATION'].ravel()
    el_coords[np.isnan(el_coords)] = -10000
    asi_azel_cal = np.stack((az_coords, el_coords), axis=-1)

    # Find the distance between the satellite azel points and
    # the asi_azel points. dist_matrix[i,j] is the distance
    # between ith asi_azel_cal value and jth sat_azel.
    dist_matrix = scipy.spatial.distance.cdist(asi_azel_cal, sat_azel, metric='euclidean')
    # Now find the minimum distance for each sat_azel.
    idx_min_dist = np.argmin(dist_matrix, axis=0)
    # if idx_min_dist == 0, it is very likely to be a NaN value
    # beacause np.argmin returns 0 if a row has a NaN.
    # NaN's arise in the first place if there is an ASI image without
    # an AC6 data point nearby in time.
    idx_min_dist = np.array(idx_min_dist, dtype=object)
    idx_min_dist[idx_min_dist == 0] = np.nan
    # For use the 1D index for the flattened ASI skymap
    # to get out the azimuth and elevation pixels.
    pixel_index = np.nan * np.ones_like(sat_azel)
    pixel_index[:, 0] = np.remainder(idx_min_dist, skymap_dict['FULL_AZIMUTH'].shape[1])
    pixel_index[:, 1] = np.floor_divide(idx_min_dist, skymap_dict['FULL_AZIMUTH'].shape[1])
    return pixel_index



import asilib.asi

import matplotlib.testing.decorators
import matplotlib.pyplot as plt

@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_geodetic_skymap'],
    tol=20,
    remove_text=True,
    extensions=['png'],
)
def test_geodetic_skymap():
    """
    Compare the (lat, lon) skymaps between the asilib and the official implementation.
    """
    time = '2020-01-01'
    ref_asi = asilib.asi.themis('GILL', time=time, load_images=False, alt=110)
    asi = asilib.asi.themis('GILL', time=time, load_images=False, custom_alt=True, alt=110)

    fig, ax = plt.subplots(2, sharex=True, sharey=True, figsize=(4, 8))
    ref_asi._pcolormesh_nan(
        ref_asi.skymap['lon'], ref_asi.skymap['lat'], ref_asi.skymap['lat'], ax[0]
        )
    asi._pcolormesh_nan(asi.skymap['lon'], asi.skymap['lat'], asi.skymap['lat'], ax[1])
    return


# TODO: Remove when cleaning up Imager
import pathlib
from datetime import datetime, timedelta
import dateutil.parser
from typing import Tuple
from copy import copy
import warnings
import re

import numpy as np
import cdflib
import scipy.io

from asilib.io.download import download_image
from asilib.io.download import download_skymap
from asilib.io import utils
import asilib


def load_image(
    asi_array_code: str,
    location_code: str,
    time: utils._time_type = None,
    time_range: utils._time_range_type = None,
    redownload: bool = False,
    time_thresh_s: float = 3,
    ignore_missing_data: bool = True,
):
    """
    Load into memory an ASI image and time stamp when ``time`` is given, or images with time stamps
    when ``time_range`` is given.

    Parameters
    ----------
    asi_array_code: str
        The imager array name, i.e. ``THEMIS`` or ``REGO``.
    location_code: str
        The ASI station code, i.e. ``ATHA``
    time: datetime.datetime or str
        The date and time to download of the data. If str, ``time`` must be in the
        ISO 8601 standard.
    time_range: list of datetime.datetimes or stings
        Defined the duration of data to download. Must be of length 2.
    redownload: bool
        If True, download the file even if it already exists. Useful if a prior
        data download was incomplete.
    time_thresh_s: float
        The maximum allowable time difference between ``time`` and an ASI time stamp.
        This is relevant only when ``time`` is specified.
    ignore_missing_data: bool
        Flag to ignore the ``FileNotFoundError`` that is raised when ASI
        data is unavailable for that date-hour. Only useful when ``time_range``
        is passed.

    Returns
    -------
    times: datetime, or List[datetime]
        The image timestamp if ``time`` is passed, or a list of timestamps if
        ``time_range`` is passed. When ``time_range`` is passed, the timestamps
        can include start time if a timestamp exactly matches, but will exclude
        the timestamp that exactly matches the end time.
    images: np.ndarray
        Either an (nPixelRows x nPixelCols) or (nTime x nPixelRows x nPixelCols)
        array containing the ASI images.

    Example
    -------
    | # Load a single image
    | asi_array_code = 'THEMIS'
    | location_code = 'ATHA'
    | time = datetime(2008, 3, 9, 9, 18, 0)
    | image_time, image = asilib.load_image(asi_array_code, location_code, time=time, redownload=False)
    |
    | # Load multiple images
    | asi_array_code = 'REGO'
    | location_code = 'LUCK'
    | time_range = [datetime(2017, 9, 27, 7, 15), datetime(2017, 9, 27, 8, 15)]
    | image_times, images = asilib.load_image(asi_array_code, location_code, time_range=time_range, redownload=False)
    """
    if (time is None) and (time_range is None):
        raise AttributeError('Neither time or time_range is specified.')
    elif (time is not None) and (time_range is not None):
        raise AttributeError('Both time and time_range can not be simultaneously specified.')

    elif time is not None:
        return _load_image(
            asi_array_code,
            location_code,
            time,
            redownload=redownload,
            time_thresh_s=time_thresh_s,
        )

    elif time_range is not None:
        return _load_images(
            asi_array_code,
            location_code,
            time_range,
            redownload=redownload,
            ignore_missing_data=ignore_missing_data,
        )
    else:
        raise ValueError("Not supposed to get here.")


def load_image_generator(
    asi_array_code: str,
    location_code: str,
    time_range: utils._time_range_type,
    redownload: bool = False,
    ignore_missing_data: bool = True,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Yields multiple ASI image files one by one and crops the time stamps by ``time_range``.

    This generator is useful for loading lots of data---useful for keograms. The returned
    time stamps span a range from time_range[0], up to, but excluding a time stamp
    exactly matching time_range[1].

    See asilib._load_images() for an example on how to use this function.

    Parameters
    ----------
    asi_array_code: str
        The imager array name, i.e. ``THEMIS`` or ``REGO``.
    location_code: str
        The ASI station code, i.e. ``ATHA``
    time_range: list of datetime.datetimes or stings
        Defined the duration of data to download. Must be of length 2.
    redownload: bool
        If True, download the file even if it already exists. Useful if a prior
        data download was incomplete.
    ignore_missing_data: bool
        Flag to ignore the ``FileNotFoundError`` that is raised when ASI
        data is unavailable for that date-hour. Only useful when ``time_range``
        is passed.

    Yields
    -------
    times: datetime
        The image timestamps contained in time_range, including the start time
        and excluding the end time (if time_range[1] exactly matches a ASI time
        stamp).
    images: np.ndarray
        An (nTime x nPixelRows x nPixelCols) array containing the ASI images
        for times contained in time_range.
    """

    time_range = utils._validate_time_range(time_range)

    # Figure out the data keys to load.
    if asi_array_code.lower() == 'rego':
        image_key = f'clg_rgf_{location_code.lower()}'
        time_key = f'clg_rgf_{location_code.lower()}_epoch'
    elif asi_array_code.lower() == 'themis':
        image_key = f'thg_asf_{location_code.lower()}'
        time_key = f'thg_asf_{location_code.lower()}_epoch'

    hours = utils._get_hours(time_range)

    for hour in hours:
        try:
            cdf_path = _find_img_path(asi_array_code, location_code, hour, redownload=redownload)
            cdf_obj = cdflib.CDF(cdf_path)
        except FileNotFoundError:
            if ignore_missing_data:
                continue
            else:
                raise

        epoch = _get_epoch(cdf_obj, time_key, hour, asi_array_code, location_code)

        idx = np.where((epoch >= time_range[0]) & (epoch < time_range[1]))[0]
        if len(idx) == 0:
            # This happens occasionally if, for example, time_range[0].minute = 57 while
            # epoch[-1].minute = 55. Continue to not try to yield an empty file.
            warnings.warn(
                f'{asi_array_code}-{location_code}, {hour} UT has no images in the time range.'
            )
            continue
        yield epoch[idx], cdf_obj.varget(image_key, startrec=idx[0], endrec=idx[-1])


def load_skymap(
    asi_array_code: str,
    location_code: str,
    time: utils._time_type,
    redownload: bool = False,
) -> dict:
    """
    Loads the appropriate THEMIS or REGO skymap file (closest and before ``time``) into memory.

    Parameters
    ----------
    asi_array_code: str
        The imager array name, i.e. ``THEMIS`` or ``REGO``.
    location_code: str
        The ASI station code, i.e. ``ATHA``
    time: datetime.datetime or str
        The date and time to download of the data. If str, ``time`` must be in the
        ISO 8601 standard.
    redownload: bool
        If True, download the file even if it already exists. Useful if a prior
        data download was incomplete.

    Returns
    -------
    dict
        The skymap data with longitudes mapped from 0->360 to to -180->180 degrees.

    Example
    -------
    | import asilib
    |
    | rego_skymap = asilib.load_skymap('REGO', 'GILL', '2018-10-01')
    """
    # Try to convert time to datetime object if it is a string.
    if isinstance(time, str):
        time = dateutil.parser.parse(time)

    if redownload:
        skymap_paths = download_skymap(asi_array_code.lower(), location_code, redownload=redownload)

    else:
        # If the user does not want to force download skymap files,
        # look for the appropriate file on the computer. If a local
        # skymap file is not found, download them all and look for the
        # appropriate file.
        skymap_dir = pathlib.Path(
            asilib.config['ASI_DATA_DIR'], asi_array_code.lower(), 'skymap', location_code.lower()
        )
        skymap_paths = sorted(
            list(skymap_dir.rglob(f'{asi_array_code.lower()}_skymap_{location_code.lower()}*'))
        )

        # Download skymap files if they are not downloaded yet.
        if len(skymap_paths) == 0:
            skymap_paths = download_skymap(
                asi_array_code.lower(), location_code, redownload=redownload
            )

    skymap_dates = _extract_skymap_dates(skymap_paths)

    # Find the skymap_date that is closest and before time.
    # For reference: dt > 0 when time is after skymap_date.
    dt = np.array([(time - skymap_date).total_seconds() for skymap_date in skymap_dates])
    dt[dt < 0] = np.inf  # Mask out all skymap_dates after time.
    if np.all(~np.isfinite(dt)):
        # Edge case when time is before the first skymap_date.
        closest_index = 0
        warnings.warn(
            f'The requested skymap time={time} for {asi_array_code}/{location_code} is before first '
            f'skymap file: {skymap_paths[0].name}. This skymap file will be used.'
        )
    else:
        closest_index = np.nanargmin(dt)
    skymap_path = skymap_paths[closest_index]

    # Load the skymap file and convert it to a dictionary.
    skymap_file = scipy.io.readsav(str(skymap_path), python_dict=True)['skymap']
    skymap_dict = {key: copy(skymap_file[key][0]) for key in skymap_file.dtype.names}

    skymap_dict = _tranform_longitude_to_180(skymap_dict)
    skymap_dict = _flip_skymap(skymap_dict)
    skymap_dict['SKYMAP_PATH'] = skymap_path
    return skymap_dict


def _extract_skymap_dates(skymap_paths):
    """
    Extract the skymap dates from each skymap_path in skymap_paths.
    """
    skymap_dates = []

    for skymap_path in sorted(skymap_paths):
        day = re.search(r'\d{8}', skymap_path.name).group(0)
        day_obj = datetime.strptime(day, "%Y%m%d")
        skymap_dates.append(day_obj)
    return skymap_dates


def _load_image(
    asi_array_code: str,
    location_code: str,
    time: utils._time_type,
    redownload: bool = False,
    time_thresh_s: float = 3,
) -> Tuple[datetime, np.ndarray]:
    """
    Gets one ASI image image given the ASI array code (THEMIS or REGO), imager location (location_code), and
    the day date-time parameters. If a file does not locally exist, this
    function will attempt to download it.

    Parameters
    ----------
    asi_array_code: str
        The imager array name, i.e. ``THEMIS`` or ``REGO``.
    location_code: str
        The ASI station code, i.e. ``ATHA``
    time: datetime.datetime or str
        The date and time to download of the data. If str, ``time`` must be in the
        ISO 8601 standard.
    time_range: list of datetime.datetimes or stings
        Defined the duration of data to download. Must be of length 2.
    redownload: bool
        If True, download the file even if it already exists. Useful if a prior
        data download was incomplete.
    time_thresh_s: float
        The maximum allowable time difference between ``time`` and an ASI time stamp.
        This is relevant only when ``time`` is specified.

    Returns
    -------
    image_time: datetime
        The image timestamp.
    image: np.ndarray
        A 2D array of the ASI image at the date-time nearest to the
        time argument.

    Raises
    ------
    AssertionError
        If a unique time stamp was not found within time_thresh_s of
        time.

    Example
    -------
    | import asilib
    |
    | time, image = asilib.io.load._load_image('2016-10-29T04:15:00', 'REGO', 'GILL')
    """
    time = utils._validate_time(time)

    cdf_path = _find_img_path(asi_array_code, location_code, time, redownload=redownload)
    cdf_obj = cdflib.CDF(cdf_path)

    if asi_array_code.lower() == 'rego':
        image_key = f'clg_rgf_{location_code.lower()}'
        time_key = f'clg_rgf_{location_code.lower()}_epoch'
    elif asi_array_code.lower() == 'themis':
        image_key = f'thg_asf_{location_code.lower()}'
        time_key = f'thg_asf_{location_code.lower()}_epoch'

    # Convert the CDF_EPOCH (milliseconds from 01-Jan-0000 00:00:00)
    # to datetime objects.
    epoch = _get_epoch(cdf_obj, time_key, time, asi_array_code, location_code)
    # Find the closest time stamp to time
    idx = np.where((epoch >= time) & (epoch < time + timedelta(seconds=time_thresh_s)))[0]
    assert len(idx) == 1, (
        f'{len(idx)} number of time stamps were found '
        f'within {time_thresh_s} seconds of {time}. '
        f'You can change the time_thresh_s kwarg to find a '
        f'time stamp further away.'
    )
    return epoch[idx[0]], cdf_obj.varget(image_key)[idx[0], :, :]


def _load_images(
    asi_array_code: str,
    location_code: str,
    time_range: utils._time_range_type,
    redownload: bool = False,
    ignore_missing_data: bool = True,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Gets multiple ASI image images given the asi_array_code (THEMIS or REGO), imager location code, and
    the time_range date-time parameters. If a file does not locally exist, this
    function will attempt to download it. The returned time stamps span a range
    from time_range[0], up to, but excluding a time stamp exactly matching time_range[1].

    Parameters
    ----------
    asi_array_code: str
        The imager array name, i.e. ``THEMIS`` or ``REGO``.
    location_code: str
        The ASI station code, i.e. ``ATHA``
    time_range: list of datetime.datetimes or stings
        Defined the duration of data to download. Must be of length 2.
    redownload: bool
        If True, download the file even if it already exists. Useful if a prior
        data download was incomplete.
    time_thresh_s: float
        The maximum allowable time difference between ``time`` and an ASI time stamp.
        This is relevant only when ``time`` is specified.
    ignore_missing_data: bool
        Flag to ignore the ``FileNotFoundError`` that is raised when ASI
        data is unavailable for that date-hour. Only useful when ``time_range``
        is passed. If no data is loaded at all, an AssertionError is raised.

    Returns
    -------
    times: datetime
        The image timestamps contained in time_range, including the start time
        and excluding the end time (if time_range[1] exactly matches a ASI time
        stamp).
    image: np.ndarray
        An (nTime x nPixelRows x nPixelCols) array containing the ASI images
        for times contained in time_range.

    Raises
    ------
    NotImplementedError
        If the image dimensions are not specified for an ASI array.
    AssertionError
        If len(time_range) != 2.
    AssertionError
        If no time stamps were found in time_range, regardless of the
        ignore_missing_data value.

    Example
    -------
    | from datetime import datetime
    |
    | import asilib
    |
    | time_range = [datetime(2016, 10, 29, 4, 15), datetime(2016, 10, 29, 4, 20)]
    | times, images = asilib.io.load._load_images('REGO', 'GILL', time_range)
    """
    times, images = _create_empty_data_arrays(asi_array_code, time_range, 'images')
    image_generator = load_image_generator(
        asi_array_code,
        location_code,
        time_range,
        redownload=redownload,
        ignore_missing_data=ignore_missing_data,
    )

    start_time_index = 0
    for file_image_times, file_images in image_generator:
        end_time_index = start_time_index + file_images.shape[0]

        images[start_time_index:end_time_index, :, :] = file_images
        times[start_time_index:end_time_index] = file_image_times

        start_time_index += file_images.shape[0]

    i_not_nan = np.where(~np.isnan(images[:, 0, 0]))[0]
    assert len(i_not_nan) > 0, (
        f'{len(i_not_nan)} number of time stamps' ' were found in time_range={time_range}'
    )
    images = images[i_not_nan, :, :]
    times = times[i_not_nan]
    return times, images


def _find_img_path(
    asi_array_code: str,
    location_code: str,
    time: utils._time_type,
    redownload: bool = False,
) -> pathlib.Path:
    """
    Returns a path to an all sky full-resolution image (THEMIS:ASF, REGO:rgf) file.
    If a file is not found locally, it will attempt to download it.

    Parameters
    ----------
    asi_array_code: str
        The imager array name, i.e. ``THEMIS`` or ``REGO``.
    location_code: str
        The ASI station code, i.e. ``ATHA``
    time: datetime.datetime or str
        The date and time to download of the data. If str, ``time`` must be in the
        ISO 8601 standard.
    redownload: bool
        If True, download the file even if it already exists. Useful if a prior
        data download was incomplete.

    Returns
    -------
    pathlib.Path
        The path to the full image file. See the implementation in
        asilib.io.load._load_image() or asilib.io.load._load_images() on
        how to use cdflib to load the image cdf files.

    Raises
    ------
    FileNotFoundError
        Catches the NotADirectoryError raised by download.py, and raises
        this FileNotFoundError that clearly conveys that the file was not
        found in the file system or online.
    ValueError
        Raised if there is an error with the file finding logic (ideally
        should not be raised).

    Example
    -------
    | import asilib
    |
    | asi_file_path = asilib._find_img_path('REGO', 'GILL', '2016-10-29T04')
    """
    time = utils._validate_time(time)

    if redownload:
        file_path = download_image(
            asi_array_code.lower(), location_code, time=time, redownload=redownload
        )[0]
    else:
        # If the user does not want to force a download, look for a file on the
        # computer. If a local file is not found, try to download one.
        search_path = pathlib.Path(asilib.config['ASI_DATA_DIR'], asi_array_code.lower())
        if asi_array_code.lower() == 'themis':
            search_pattern = f'*asf*{location_code.lower()}*{time.strftime("%Y%m%d%H")}*'
        elif asi_array_code.lower() == 'rego':
            search_pattern = f'*rgf*{location_code.lower()}*{time.strftime("%Y%m%d%H")}*'
        matched_paths = list(search_path.rglob(search_pattern))

        if len(matched_paths) == 1:  # A local file found
            file_path = matched_paths[0]

        elif len(matched_paths) == 0:  # No local file found
            try:
                file_path = download_image(
                    asi_array_code.lower(), location_code, time=time, redownload=redownload
                )[0]
            except NotADirectoryError:
                raise FileNotFoundError(
                    f'{asi_array_code.upper()} ASI data not found for location_code={location_code} at {time}'
                )
        else:  # Multiple files found?
            raise ValueError(f"Not sure what happend here. I found {matched_paths} matching paths.")

    return file_path


def _get_epoch(cdf_obj, time_key, hour_date_time, asi_array_code, location_code):
    """
    Gets the CDF epoch array and modifies a ValueError when a CDF file is corrupted.
    """
    try:
        epoch = np.array(cdflib.cdfepoch.to_datetime(cdf_obj.varget(time_key)))
    except ValueError as err:
        if ('read length must be non-negative or -1' in str(err)) or ('not found' in str(err)):
            raise ValueError(
                str(err) + '\n\n ASI data is probably corrupted for '
                f'time={hour_date_time}, asi_array_code={asi_array_code}, '
                f'location_code={location_code}. This can happen when the '
                f'download was interrupted. Try to redownload the data using '
                f'redownload=True).'
            )
        else:
            raise
    return epoch


def _create_empty_data_arrays(asi_array_code, time_range, type):
    """
    Creates two appropriately sized np.arrays full of np.nan. The first is a 1d times array,
    and the second is either: a 2d array (n_steps, n_pixels) if type=='keogram', or a 3d array
    (n_times, n_pixels, n_pixels) if type='images'.
    """
    if asi_array_code.lower() == 'themis':
        img_size = 256
        cadence_s = 3
    elif asi_array_code.lower() == 'rego':
        img_size = 512
        cadence_s = 3
    else:
        raise NotImplementedError

    time_range = utils._validate_time_range(time_range)
    # + 1 just in case. The THEMIS and REGO cadence is not
    # always exactly 3 seconds, so in certain circumstances
    # there may be one or two extra data points.
    max_n_timestamps = int((time_range[1] - time_range[0]).total_seconds() / cadence_s) + 1

    if type.lower() == 'keogram':
        data_shape = (max_n_timestamps, img_size)
    elif type.lower() == 'images':
        data_shape = (max_n_timestamps, img_size, img_size)
    else:
        raise ValueError('type must be "keogram" or "images".')

    # object is the only dtype that can contain datetime objects
    times = np.nan * np.zeros(max_n_timestamps, dtype=object)
    data = np.nan * np.zeros(data_shape)
    return times, data


def _flip_skymap(skymap):
    """
    IDL is a column-major language while Python is row-major. This function
    tranposes the 2- and 3-D arrays to make them compatable with the images
    that are saved in row-major.
    """
    for key in skymap:
        if hasattr(skymap[key], 'shape'):
            shape = skymap[key].shape
            if (len(shape) == 2) and (shape[0] == shape[1]):
                skymap[key] = skymap[key][::-1, ::-1]  # For Az/El maps.
            elif (len(shape) == 3) and (shape[1] == shape[2]):
                skymap[key] = skymap[key][:, ::-1, ::-1]  # For lat/lon maps
    return skymap


def _tranform_longitude_to_180(skymap):
    """
    Transform the SITE_MAP_LONGITUDE and FULL_MAP_LONGITUDE arrays from
    (0 -> 360) to (-180 -> 180).
    """
    skymap['SITE_MAP_LONGITUDE'] = np.mod(skymap['SITE_MAP_LONGITUDE'] + 180, 360) - 180

    # Don't take the modulus of NaNs
    valid_val_idx = np.where(~np.isnan(skymap['FULL_MAP_LONGITUDE']))
    skymap['FULL_MAP_LONGITUDE'][valid_val_idx] = (
        np.mod(skymap['FULL_MAP_LONGITUDE'][valid_val_idx] + 180, 360) - 180
    )
    return skymap



# TODO: Remove when cleaning up Imager
"""
A set of utility functions for asilib.
"""
import dateutil.parser
import pathlib
from typing import List, Union
import copy
from datetime import timedelta, datetime

from bs4 import BeautifulSoup
import requests
import numpy as np


_time_type = Union[datetime, str]
_time_range_type = List[_time_type]


def _validate_time(time: _time_type) -> List[datetime]:
    """
    Validates tries to parse the time into datetime objects.
    """
    if isinstance(time, str):
        time = dateutil.parser.parse(time)
    elif isinstance(time, (int, float)):
        raise ValueError(f'Unknown time format, {time}')
    _is_valid_year(time)
    return time


def _validate_time_range(time_range: _time_range_type) -> List[datetime]:
    """
    Validates tries to parse the time_range into datetime objects.
    """
    if time_range is None:
        return None

    assert isinstance(
        time_range, (list, tuple, np.ndarray)
    ), "time_range must be a list, tuple, or np.ndarray."
    assert len(time_range) == 2, "time_range must be a list or a tuple with start and end times."

    time_range_parsed = []

    for t in time_range:
        if isinstance(t, str):
            time_range_parsed.append(dateutil.parser.parse(t))
        elif isinstance(t, (int, float)):
            raise ValueError(f'Unknown time format, {t}')
        else:
            time_range_parsed.append(t)

    for t in time_range_parsed:
        _is_valid_year(t)

    time_range_parsed.sort()
    return time_range_parsed


def _is_valid_year(time):
    """
    Provides a sanity check that the year is after 2000 and before the current year + 1.
    """
    year = time.year
    if year < 2000:
        raise ValueError(f'The passed year={year} must be greater than 2000.')
    elif year > datetime.now().year + 1:
        raise ValueError(f'The passed year={year} must be less than the current year + 1.')
    return


def _get_hours(time_range: _time_range_type) -> List[datetime]:
    """
    Helper function to figure out what date-hour times are between the times in time_range.
    This function is useful to figure out what hourly ASI files to download.
    """
    # Modify time_range. If time_range[0] is not at the top of the hour, we zero the minutes
    # seconds, and milliseconds. This helps with keeping the + 1 hour offsets aligned to the
    # start of the hour.
    current_hour = copy.copy(time_range[0].replace(minute=0, second=0, microsecond=0))
    hours = []

    # Not <= in while loop because we don't want to download the final hour if time_range[1] is,
    # for example, 05:00:00 [HH:MM:SS].
    while current_hour < time_range[1]:
        hours.append(current_hour)
        current_hour += timedelta(hours=1)
    return hours


def _stream_large_file(
    url: str, save_path: Union[pathlib.Path, str], test_flag: bool = False
) -> None:
    """
    Streams a file from url to save_path. In requests.get(), stream=True
    sets up a generator to download a small chuck of data at a time,
    instead of downloading the entire file into RAM first.

    Parameters
    ----------
    url: str
        The URL to the file.
    save_path: str or pathlib.Path
        The local save path for the file.
    test_flag: bool (optional)
        If True, the download will halt after one 5 Mb chunk of data is
        downloaded.

    Returns
    -------
    None
    """
    r = requests.get(url, stream=True)
    file_size = int(r.headers.get('content-length'))
    downloaded_bites = 0

    save_name = pathlib.Path(save_path).name

    megabyte = 1024 * 1024

    with open(save_path, 'wb') as f:
        for data in r.iter_content(chunk_size=5 * megabyte):
            f.write(data)
            if test_flag:
                return
            # Update the downloaded % in the terminal.
            downloaded_bites += len(data)
            download_percent = round(100 * downloaded_bites / file_size)
            download_str = "#" * (download_percent // 5)
            print(f'Downloading {save_name}: |{download_str:<20}| {download_percent}%', end='\r')
    print()  # Add a newline
    return


def _search_hrefs(url: str, search_pattern: str = '.cdf') -> List[str]:
    """
    Given a url string, this function returns all of the
    hyper references (hrefs, or hyperlinks). If search_pattern is not
    specified, a default '.cdf' value is assumed and this function
    will return all hrefs with the CDF extension. If no hrefs containing
    search_pattern are found, this function raises a NotADirectoryError.
    The search is case-insensitive.

    Parameters
    ----------
    url: str
        A url in string format
    search_pattern: str (optional)
        Find the exact search_pattern text contained in the hrefs.
        By default all hrefs matching the extension ".cdf" are returned.

    Returns
    -------
    hrefs: List(str)
        A list of hrefs that contain the search_pattern.

    Raises
    ------
    NotADirectoryError
        If a hyper reference (a folder or a file) is not found on the
        server. This is raised if the data does not exist.
    """
    matched_hrefs = []

    request = requests.get(url)
    # request.status_code
    soup = BeautifulSoup(request.content, 'html.parser')

    for href in soup.find_all('a', href=True):
        if search_pattern.lower() in href['href'].lower():
            matched_hrefs.append(href['href'])
    if len(matched_hrefs) == 0:
        raise NotADirectoryError(
            f'The url {url} does not contain any hyper '
            f'references containing the search_pattern="{search_pattern}".'
        )
    return matched_hrefs



from typing import List
import warnings

import matplotlib.pyplot as plt
import matplotlib.colors as colors
import numpy as np

import asilib.io.utils as utils
from asilib.analysis.keogram import keogram


def plot_keogram(
    asi_array_code: str,
    location_code: str,
    time_range: utils._time_range_type,
    map_alt: float = None,
    path: np.array = None,
    aacgm: bool = False,
    ax: plt.Axes = None,
    color_bounds: List[float] = None,
    color_norm: str = 'lin',
    title: bool = True,
    pcolormesh_kwargs: dict = {},
):
    """
    Makes a keogram along the central meridian.

    .. warning::
        Use :py:meth:`~asilib.imager.Imager.plot_keogram()` instead. This function will be
        removed in or after December 2023.

    Parameters
    ----------
    asi_array_code: str
        The imager array name, i.e. ``THEMIS`` or ``REGO``.
    location_code: str
        The ASI station code, i.e. ``ATHA``
    time_range: list of datetime.datetimes or stings
        Defined the duration of data to download. Must be of length 2.
    map_alt: int
        The mapping altitude, in kilometers, used to index the mapped latitude in the
        skymap calibration data. If None, will plot pixel index for the y-axis.
    path: array
        Make a keogram along a custom path. Path shape must be (n, 2) and contain the
        lat/lon coordinates that are mapped to map_alt. If the map_alt kwarg is
        unspecified, this function will raise a ValueError.
    aacgm: bool
        Map the keogram latitudes to Altitude Adjusted Corrected Geogmagnetic Coordinates
        (aacgmv2) derived by Shepherd, S. G. (2014), Altitude-adjusted corrected geomagnetic
        coordinates: Definition and functional approximations, Journal of Geophysical
        Research: Space Physics, 119, 7501-7521, doi:10.1002/2014JA020264.
    ax: plt.Axes
        The subplot to plot the image on. If None, this function will
        create one.
    color_bounds: List[float]
        The lower and upper values of the color scale. If None, will
        automatically set it to low=1st_quartile and
        high=min(3rd_quartile, 10*1st_quartile)
    color_norm: str
        Sets the linear ('lin') or logarithmic ('log') color normalization.
    title: bool
        Toggles a default plot title with the format "date ASI_array_code-location_code keogram".
    pcolormesh_kwargs: dict
        A dictionary of keyword arguments (kwargs) to pass directly into
        plt.pcolormesh. One use of this parameter is to change the colormap. For example,
        pcolormesh_kwargs = {'cmap':'tu'}

    Returns
    -------
    ax: plt.Axes
        The subplot object to modify the axis, labels, etc.
    im: plt.AxesImage
        The plt.pcolormesh image object. Common use for im is to add a colorbar.

    Raises
    ------
    AssertionError
        If len(time_range) != 2. Also if map_alt does not equal the mapped
        altitudes in the skymap mapped values.

    Example
    -------
    | import matplotlib.pyplot as plt
    |
    | import asilib
    |
    | asi_array_code='REGO'
    | location_code='LUCK'
    | time_range=['2017-09-27T07', '2017-09-27T09']
    |
    | fig, ax = plt.subplots(figsize=(8, 6))
    | ax, im = asilib.plot_keogram(asi_array_code, location_code, time_range,
    |                ax=ax, map_alt=230, color_bounds=(300, 800), pcolormesh_kwargs={'cmap':'turbo'})
    |
    | plt.colorbar(im)
    | plt.tight_layout()
    | plt.show()
    """
    warnings.warn(
        "Use asilib.Imager.plot_keogram() instead. This function will be removed "
        "in or after December 2023.",
        DeprecationWarning,
    )
    time_range = utils._validate_time_range(time_range)
    keo_df = keogram(asi_array_code, location_code, time_range, map_alt, path=path, aacgm=aacgm)

    if ax is None:
        _, ax = plt.subplots()

    # Figure out the color_bounds from the image data.
    if color_bounds is None:
        lower, upper = np.quantile(keo_df, (0.25, 0.98))
        color_bounds = [lower, np.min([upper, lower * 10])]

    if color_norm == 'log':
        norm = colors.LogNorm(vmin=color_bounds[0], vmax=color_bounds[1])
    elif color_norm == 'lin':
        norm = colors.Normalize(vmin=color_bounds[0], vmax=color_bounds[1])
    else:
        raise ValueError('color_norm must be either "log" or "lin".')

    im = ax.pcolormesh(
        keo_df.index,
        keo_df.columns,
        keo_df.to_numpy()[:-1, :-1].T,
        norm=norm,
        shading='flat',
        **pcolormesh_kwargs,
    )

    if title:
        ax.set_title(
            f'{time_range[0].date()} | {asi_array_code.upper()}-{location_code.upper()} keogram'
        )
    return ax, im



import numpy as np
import pandas as pd
import warnings
import aacgmv2

from asilib.io import utils
from asilib.io.load import (
    load_image_generator,
    load_skymap,
    _create_empty_data_arrays,
)


def keogram(
    asi_array_code: str,
    location_code: str,
    time_range: utils._time_range_type,
    map_alt: int = None,
    path: np.array = None,
    aacgm: bool = False,
):
    """
    Makes a keogram pd.DataFrame along the central meridian.

    Parameters
    ----------
    asi_array_code: str
        The imager array name, i.e. ``THEMIS`` or ``REGO``.
    location_code: str
        The ASI station code, i.e. ``ATHA``
    time_range: list of datetime.datetimes or stings
        Defined the duration of data to download. Must be of length 2.
    map_alt: int
        The mapping altitude, in kilometers, used to index the mapped latitude in the
        skymap data. If None, will plot pixel index for the y-axis.
    path: array
        Make a keogram along a custom path. Path shape must be (n, 2) and contain the
        lat/lon coordinates that are mapped to map_alt. If the map_alt kwarg is
        unspecified, this function will raise a ValueError.
    aacgm: bool
        Map the keogram latitudes to Altitude Adjusted Corrected Geogmagnetic Coordinates
        (aacgmv2) derived by Shepherd, S. G. (2014), Altitude-adjusted corrected geomagnetic
        coordinates: Definition and functional approximations, Journal of Geophysical
        Research: Space Physics, 119, 7501-7521, doi:10.1002/2014JA020264.

    Returns
    -------
    keo: pd.DataFrame
        The 2d keogram with the time index. The columns are the geographic latitude
        if map_alt != None, otherwise it is the image pixel values (0-265) or (0-512).

    Raises
    ------
    AssertionError
        If map_alt does not equal the mapped altitudes in the skymap mapped values.
    ValueError
        If no images are in ``time_range``.
    ValueError
        If a custom path is provided but not map_alt.
    """
    keo = Keogram(asi_array_code, location_code, time_range)
    return keo.keogram(map_alt, path, aacgm)


def ewogram(
    asi_array_code: str, location_code: str, time_range: utils._time_range_type, map_alt: int = None
):
    """
    Makes a East-West ewogram.

    Parameters
    ----------
    asi_array_code: str
        The imager array name, i.e. ``THEMIS`` or ``REGO``.
    location_code: str
        The ASI station code, i.e. ``ATHA``
    time_range: list of datetime.datetimes or stings
        Defined the duration of data to download. Must be of length 2.
    map_alt: int
        The mapping altitude, in kilometers, used to index the mapped longitude in the
        skymap data. If None, will plot pixel index for the y-axis.

    Returns
    -------
    ewo: pd.DataFrame
        The 2d ewogram with the time index. The columns are the geographic longitude
        if map_alt != None, otherwise it is the image pixel indices.

    Raises
    ------
    AssertionError
        If map_alt does not equal the mapped altitudes in the skymap mapped values.
    ValueError
        If no imager data is found in ``time_range``.
    """
    # TODO: Add tests for an ewogram.
    image_generator = load_image_generator(asi_array_code, location_code, time_range)
    ewo_times, ewo = _create_empty_data_arrays(asi_array_code, time_range, 'keogram')

    start_time_index = 0
    for file_image_times, file_images in image_generator:
        end_time_index = start_time_index + file_images.shape[0]
        ewo[start_time_index:end_time_index, :] = file_images[:, ewo.shape[1] // 2, :]

        ewo_times[start_time_index:end_time_index] = file_image_times
        start_time_index += file_images.shape[0]

    # This code block removes any filler nan values if the ASI images were not sampled at the instrument
    # cadence throughout time_range.
    i_valid = np.where(~np.isnan(ewo[:, 0]))[0]
    ewo = ewo[i_valid, :]
    ewo_times = ewo_times[i_valid]

    if not ewo.shape[0]:
        raise ValueError(
            f'The keogram is empty for {asi_array_code}/{location_code} '
            f'during {time_range}. The image data probably does not exist '
            f'in this time interval'
        )

    if map_alt is None:
        ewo_longitude = np.arange(ewo.shape[0])  # Dummy index values for longitudes.
    else:
        skymap = load_skymap(asi_array_code, location_code, time_range[0])
        assert (
            map_alt in skymap['FULL_MAP_ALTITUDE'] / 1000
        ), f'{map_alt} km is not in skymap altitudes: {skymap["FULL_MAP_ALTITUDE"]/1000} km'
        alt_index = np.where(skymap['FULL_MAP_ALTITUDE'] / 1000 == map_alt)[0][0]
        ewo_longitude = skymap['FULL_MAP_LONGITUDE'][alt_index, ewo.shape[0] // 2, :]

        # ewo_longitude array are at the pixel edges. Remap it to the centers
        dl = ewo_longitude[1:] - ewo_longitude[:-1]
        ewo_longitude = ewo_longitude[0:-1] + dl / 2

        # Since keogram_latitude values are NaNs near the image edges, we want to filter
        # out those indices from keogram_latitude and keo.
        valid_lons = np.where(~np.isnan(ewo_longitude))[0]
        ewo_longitude = ewo_longitude[valid_lons]
        keo = ewo[:, valid_lons]
    return pd.DataFrame(data=keo, index=ewo_times, columns=ewo_longitude)


class Keogram:
    # TODO: Change the asi args to asi_info dictionary that is created by asilib.themis() or asilib.rego() functions.
    def __init__(
        self, asi_array_code: str, location_code: str, time_range: utils._time_range_type
    ) -> None:
        """
        Keogram class that makes keograms and ewograms along the meridian or a custom path.

        Parameters
        ----------
        asi_array_code: str
            The imager array name, i.e. ``THEMIS`` or ``REGO``.
        location_code: str
            The ASI station code, i.e. ``ATHA``
        time_range: list of datetime.datetimes or stings
            Defined the duration of data to download. Must be of length 2.
        """
        # __init__ saves the variables needed to get the images.
        self.asi_array_code = asi_array_code
        self.location_code = location_code
        self.time_range = utils._validate_time_range(time_range)

        # In the full implementation we won't need these if-else statements.
        if self.asi_array_code.lower() == 'themis':
            self.img_size = 256
            self.asi_cadence_s = 3
        elif self.asi_array_code.lower() == 'rego':
            self.img_size = 512
            self.asi_cadence_s = 3
        else:
            raise NotImplementedError
        return

    def keogram(self, map_alt: int = None, path: np.array = None, aacgm=False):
        """
        Creates a keogram along the meridian or a custom path.

        Parameters
        ----------
        map_alt: int
            The mapping altitude, in kilometers, used to index the mapped latitude in the
            skymap data. If None, will plot pixel index for the y-axis.
        path: array
            Make a keogram along a custom path. Path shape must be (n, 2) and contain the
            lat/lon coordinates that are mapped to map_alt. If the map_alt kwarg is
            unspecified, this function will raise a ValueError.
        aacgm: bool
            Map the keogram latitudes to Altitude Adjusted Corrected Geogmagnetic Coordinates
            (aacgmv2) derived by Shepherd, S. G. (2014), Altitude-adjusted corrected geomagnetic
            coordinates: Definition and functional approximations, Journal of Geophysical
            Research: Space Physics, 119, 7501-7521, doi:10.1002/2014JA020264.

        Returns
        -------
        keo: pd.DataFrame
            The 2d keogram with the time index. The columns are the (geographic or magnetic)
            latitudes if map_alt is specified, otherwise it is the image pixel indices.

        Raises
        ------
        AssertionError
            If map_alt does not equal the mapped altitudes in the skymap mapped values.
        ValueError
            If no images are in ``time_range``.
        ValueError
            If a custom path is provided but not map_alt.
        """
        self.map_alt = map_alt
        self.aacgm = aacgm
        self.path = path
        keo_times, self._keo = self._empty_keogram(self.time_range)
        self.skymap = load_skymap(self.asi_array_code, self.location_code, self.time_range[0])

        # Determine what pixels to slice
        self._keogram_pixels()
        # Not all of the pixels are valid (e.g. below the horizon)
        self._keo = self._keo[:, : self._pixels.shape[0]]

        self._keogram_latitude()

        # Load and slice images.
        image_generator = load_image_generator(
            self.asi_array_code, self.location_code, self.time_range
        )
        start_time_index = 0
        for file_image_times, file_images in image_generator:
            # Check if the image time stamps are out of order. Otherwise this will cause
            # a broadcast numpy error.
            dt = np.array(
                [
                    (f - i).total_seconds()
                    for i, f in zip(file_image_times[:-1], file_image_times[1:])
                ]
            )
            if not np.all(dt > 0):
                raise ValueError(
                    f'{self.asi_array_code}-{self.location_code} time stamps are out of order.'
                )
            end_time_index = start_time_index + file_images.shape[0]
            self._keo[start_time_index:end_time_index, :] = file_images[
                :, self._pixels[:, 0], self._pixels[:, 1]
            ]
            keo_times[start_time_index:end_time_index] = file_image_times
            start_time_index += file_images.shape[0]

        # Remove NaN times.
        i_valid = np.where(~np.isnan(self._keo[:, 0]))[0]
        keo = self._keo[i_valid, :]
        keo_times = keo_times[i_valid]

        if not keo.shape[0]:
            raise ValueError(
                f'The keogram is empty for {self.asi_array_code}/{self.location_code} '
                f'during {self.time_range}. The image data probably does not exist '
                f'in this time interval'
            )
        self.keo = pd.DataFrame(data=keo, index=keo_times, columns=self.keogram_latitude)
        return self.keo

    def _empty_keogram(self, time_range):
        """
        Creates an empty 2D keogram and time arrays.
        """
        time_range = utils._validate_time_range(time_range)
        # + 1 just in case. The THEMIS and REGO cadence is not
        # always exactly 3 seconds, so in certain circumstances
        # there may be one or two extra data points.
        max_n_timestamps = (
            int((time_range[1] - time_range[0]).total_seconds() / self.asi_cadence_s) + 1
        )
        data_shape = (max_n_timestamps, self.img_size)

        # object is the only dtype that can contain datetime objects
        times = np.nan * np.zeros(max_n_timestamps, dtype=object)
        data = np.nan * np.zeros(data_shape)
        return times, data

    def _keogram_pixels(self, minimum_elevation=0):
        """
        Find what pixels to index and reshape the keogram.
        """
        # CASE 1: No path provided. Output self._pixels that slice the meridian.
        if self.path is None:
            self._pixels = np.column_stack(
                (
                    np.arange(self._keo.shape[1]),
                    self._keo.shape[1] * np.ones(self._keo.shape[1]) // 2,
                )
            ).astype(int)

        # CASE 2: A path is provided so now we need to calculate the custom path
        # on the lat/lon skymap.
        else:
            if (self.map_alt is None) or (
                self.map_alt not in self.skymap['FULL_MAP_ALTITUDE'] / 1000
            ):
                raise ValueError(
                    f'{self.map_alt} km is not in skymap altitudes: {self.skymap["FULL_MAP_ALTITUDE"]/1000} km'
                )
            alt_index = np.where(self.skymap['FULL_MAP_ALTITUDE'] / 1000 == self.map_alt)[0][0]
            self._pixels = self._path_to_pixels(self.path, alt_index)

        above_elevation = np.where(
            self.skymap['FULL_ELEVATION'][self._pixels[:, 0], self._pixels[:, 1]]
            >= minimum_elevation
        )[0]
        self._pixels = self._pixels[above_elevation]
        return

    def _path_to_pixels(self, path, alt_index, threshold=1):
        """
        Convert the lat/lon path that is mapped to map_alt in kilometers to
        the x- and y-pixels in the skymap lat/lon mapped file.

        Parameters
        ----------
        path: np.array
            The lat/lon array of shape (n, 2).
        alt_index: int
            Determines what altitude index to use.
        threshold: float
            The maximum distance threshold, in degrees, between the path (lat, lon)
            and the skymap (lat, lon)

        Returns
        -------
        np.array
            The valid x pixels corresponding to the path.
        np.array
            The valid y pixels corresponding to the path.
        np.array
            Path indices corresponding to rows with a pixel within threshold
            degrees distance.
        """
        if np.array(path).shape[0] == 0:
            raise ValueError('The path is empty.')
        if np.any(np.isnan(path)):
            raise ValueError("The lat/lon path can't contain NaNs.")
        if np.any(np.max(path) > 180) or np.any(np.min(path) < -180):
            raise ValueError("The lat/lon values must be in the range -180 to 180.")

        nearest_pixels = np.nan * np.zeros_like(path)

        for i, (lat, lon) in enumerate(path):
            distances = np.sqrt(
                (self.skymap['FULL_MAP_LATITUDE'][alt_index, :, :] - lat) ** 2
                + (self.skymap['FULL_MAP_LONGITUDE'][alt_index, :, :] - lon) ** 2
            )
            idx = np.where(distances == np.nanmin(distances))

            # Keep NaNs if distanace is larger than threshold.
            if distances[idx][0] > threshold:
                warnings.warn(
                    f'Some of the keogram path coordinates are outside of the '
                    f'maximum {threshold} degrees distance from the nearest '
                    f'skymap map coordinate.'
                )
                continue
            nearest_pixels[i, :] = [idx[0][0], idx[1][0]]

        valid_pixels = np.where(np.isfinite(nearest_pixels[:, 0]))[0]
        if valid_pixels.shape[0] == 0:
            raise ValueError('The keogram path is completely outside of the skymap.')
        return nearest_pixels[valid_pixels, :].astype(int)

    def _keogram_latitude(self):
        """
        Keogram's vertical axis: geographic latitude, magnetic latitude, or pixel index.
        """
        if self.map_alt is None:
            self.keogram_latitude = np.arange(self._keo.shape[1])
        else:
            alt_indices = np.where(self.skymap['FULL_MAP_ALTITUDE'] / 1000 == self.map_alt)[0]
            assert len(alt_indices) == 1, (
                f"{self.map_alt} is not in the skymap altitudes: "
                f"{self.skymap['FULL_MAP_ALTITUDE'] / 1000}"
            )
            self.keogram_latitude = self.skymap['FULL_MAP_LATITUDE'][
                alt_indices[0], self._pixels[:, 0], self._pixels[:, 1]
            ]
            if self.aacgm:
                longitudes = self.skymap['FULL_MAP_LONGITUDE'][
                    alt_indices[0], self._pixels[:, 0], self._pixels[:, 1]
                ]
                self.keogram_latitude = aacgmv2.convert_latlon_arr(
                    self.keogram_latitude,
                    longitudes,
                    self.map_alt,
                    self.time_range[0],
                    method_code="G2A",
                )[0]
        return



import numpy as np
import pandas as pd
import warnings

import asilib

earth_radius_km = 6371  # Earth radius


def equal_area(asi_array_code, location_code, time, lla, box_km=(5, 5), alt_thresh_km=3):
    """
    Calculate a ``box_km`` area at the aurora emission altitude.

    Parameters
    ----------
    asi_array_code: str
        The imager array name, i.e. ``THEMIS`` or ``REGO``.
    location_code: str
        The ASI station code, i.e. ``ATHA``
    time: datetime.datetime or str
        The date and time to download of the data. If str, ``time`` must be in the
        ISO 8601 standard.
    lla: np.ndarray
        An array with (n_time, 3) dimensions with the columns
        representing the latitude, longitude, and altitude (LLA)
        coordinates.
    box_size_km: iterable
        Bounds the emission box dimensions in longitude and latitude. Units are kilometers.

    Returns
    -------
    pixel_mask: np.ndarray
        An array with (n_time, n_x_pixels, n_y_pixels) dimensions with
        dimensions n_x_pixels and n_y_pixels dimensions the size of each
        image. Values inside the area are 1 and outside are np.nan.
    """
    assert len(box_km) == 2, 'The box_km parameter must have a length of 2.'

    skymap_dict = asilib.load_skymap(asi_array_code, location_code, time)

    # Get numpy array if pd.DataFrame passed
    if isinstance(lla, pd.DataFrame):
        lla = lla.to_numpy()
    elif isinstance(lla, list):
        lla = np.array(lla)

    initial_shape = lla.shape
    # I want to work with 2d arrays, even if 1d array was supplied. We will reduce the dimensions
    # at the end.
    if len(initial_shape) == 1:
        lla = np.array([lla])

    # Check that the altitude value is in the skymap.
    for alt in lla[:, -1]:
        assert (
            np.min(np.abs(skymap_dict['FULL_MAP_ALTITUDE'] / 1000 - alt)) < alt_thresh_km
        ), f'Got {alt} km altitude, but it must be one of these: {skymap_dict["FULL_MAP_ALTITUDE"]/1000}'
    alt_index = np.argmin(np.abs(skymap_dict['FULL_MAP_ALTITUDE'] / 1000 - alt))
    lat_map = skymap_dict['FULL_MAP_LATITUDE'][alt_index, :, :]
    lon_map = skymap_dict['FULL_MAP_LONGITUDE'][alt_index, :, :]

    # shape[X]-1 because the lat/lon maps define the vertices.
    pixel_mask = np.nan * np.zeros((lla.shape[0], lat_map.shape[0] - 1, lat_map.shape[1] - 1))

    dlat = _dlat(box_km[1], lla[:, -1])
    dlon = _dlon(box_km[0], lla[:, -1], lla[:, 0])

    for i, ((lat, lon, _), dlon_i, dlat_i) in enumerate(zip(lla, dlon, dlat)):
        # Check that the lat/lon values are inside the skymap coordinates.
        if (
            (lat > np.nanmax(lat_map))
            or (lat < np.nanmin(lat_map))
            or (lon > np.nanmax(lon_map))
            or (lon < np.nanmin(lon_map))
        ):
            warnings.warn(
                'Some latitude or longitude values are outside of the skymap '
                'lat/lon arrays. The equal area mask will be all NaNs.'
            )
            continue

        # Find the indices of the box. If none were found (pixel smaller than
        # the box_size_km) then increase the box size until one pixel is found.
        masked_box_len = 0
        multiplier = 1
        step = 0.1

        while masked_box_len == 0:
            idx_box = np.where(
                (lat_map >= lat - multiplier * dlat_i / 2)
                & (lat_map <= lat + multiplier * dlat_i / 2)
                & (lon_map >= lon - multiplier * dlon_i / 2)
                & (lon_map <= lon + multiplier * dlon_i / 2)
            )

            masked_box_len = len(idx_box[0])
            if masked_box_len:
                pixel_mask[i, idx_box[0], idx_box[1]] = 1
            else:
                multiplier += step

    if len(initial_shape) == 1:
        return pixel_mask.reshape(lat_map.shape)
    else:
        return pixel_mask


def _dlat(d, alt):
    """
    Calculate the change in latitude that correpsponds to arc length distance d at
    alt altitude. Units are kilometers. Both d and alt must be the same length.

    Parameters
    ----------
    d: float or np.ndarray
        A float, 1d list, or 1d np.array of arc length.
    alt: float or np.ndarray
        A float, 1d list, or 1d np.array of satellite altitudes.

    Returns
    -------
    dlat: float or np.ndarray
        A float or an array of the corresponding latitude differences in degrees.
    """
    if isinstance(alt, list):  # Don't need to cast d since it is in np.divide().
        alt = np.array(alt)
    return np.rad2deg(np.divide(d, (earth_radius_km + alt)))


def _dlon(d, alt, lat):
    """
    Calculate the change in longitude that corresponds to arc length distance d at
    a lat latitude, and alt altitude.

    Parameters
    ----------
    d: float or np.ndarray
        A float or an array of arc length in kilometers.
    alt: float or np.ndarray
        A float or an array of satellite altitudes in kilometers.
    lat: float
        The latitude to evaluate the change in longitude in degrees.

    Returns
    -------
    dlon: float or np.ndarray
        A float or an array of the corresponding longitude differences in degrees.
    """
    if isinstance(alt, list):  # Don't need to cast other variables.
        alt = np.array(alt)

    numerator = np.sin(d / (2 * (earth_radius_km + alt)))
    denominator = np.cos(np.deg2rad(lat))
    dlon_rads = 2 * np.arcsin(numerator / denominator)
    return np.rad2deg(dlon_rads)
